# pylint: disable=too-many-lines
from typing import List, Tuple, Type

from tealer.teal.instructions import instructions
from tealer.teal.instructions import transaction_field
from tealer.detectors.abstract_detector import AbstractDetector
from tealer.detectors.all_detectors import (
    IsUpdatable,
    IsDeletable,
    CanCloseAccount,
    CanCloseAsset,
    AnyoneCanUpdate,
    AnyoneCanDelete,
)

from tests.utils import construct_cfg


CAN_UPDATE = """
#pragma version 2
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    txn OnCompletion
    int NoOp
    ==
    bnz handle_noop
    txn OnCompletion
    int OptIn
    ==
    bnz handle_optin
    txn OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    txn OnCompletion
    int DeleteApplication
    ==
    bnz handle_deleteapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_deleteapp:
err
"""

ins_list = [
    instructions.Pragma(2),
    instructions.Int(0),
    instructions.Txn(transaction_field.ApplicationID()),
    instructions.Eq(),
    instructions.BZ("not_creation"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("not_creation"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("NoOp"),
    instructions.Eq(),
    instructions.BNZ("handle_noop"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("OptIn"),
    instructions.Eq(),
    instructions.BNZ("handle_optin"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("CloseOut"),
    instructions.Eq(),
    instructions.BNZ("handle_closeout"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("DeleteApplication"),
    instructions.Eq(),
    instructions.BNZ("handle_deleteapp"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("handle_noop"),
    instructions.Label("handle_optin"),
    instructions.Label("handle_closeout"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("handle_deleteapp"),
    instructions.Err(),
]

ins_partitions = [
    (0, 5),
    (5, 7),
    (7, 12),
    (12, 16),
    (16, 20),
    (20, 24),
    (24, 26),
    (26, 27),
    (27, 28),
    (28, 31),
    (31, 33),
]

bbs_links = [
    (0, 1),
    (0, 2),
    (2, 3),
    (2, 7),
    (3, 4),
    (3, 8),
    (4, 5),
    (4, 9),
    (5, 6),
    (5, 10),
    (7, 8),
    (8, 9),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

CAN_UPDATE_VULNERABLE_PATHS = [
    [bbs[0], bbs[2], bbs[3], bbs[4], bbs[5], bbs[6]],
]

CAN_UPDATE_LOOP = """
#pragma version 4
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    txn OnCompletion
    int NoOp
    ==
    bnz handle_noop
    txn OnCompletion
    int OptIn
    ==
    bnz handle_optin
    txn OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    txn OnCompletion
    int DeleteApplication
    ==
    bnz handle_deleteapp
    b loop
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_deleteapp:
err
loop:
    load 0
    int 5
    <
    bz end
    load 0
    int 1
    +
    store 0
    b loop
end:
    int 1
    return
"""

ins_list = [
    instructions.Pragma(4),
    instructions.Int(0),
    instructions.Txn(transaction_field.ApplicationID()),
    instructions.Eq(),
    instructions.BZ("not_creation"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("not_creation"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("NoOp"),
    instructions.Eq(),
    instructions.BNZ("handle_noop"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("OptIn"),
    instructions.Eq(),
    instructions.BNZ("handle_optin"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("CloseOut"),
    instructions.Eq(),
    instructions.BNZ("handle_closeout"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("DeleteApplication"),
    instructions.Eq(),
    instructions.BNZ("handle_deleteapp"),
    instructions.B("loop"),
    instructions.Label("handle_noop"),
    instructions.Label("handle_optin"),
    instructions.Label("handle_closeout"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("handle_deleteapp"),
    instructions.Err(),
    instructions.Label("loop"),
    instructions.Load(0),
    instructions.Int(5),
    instructions.Less(),
    instructions.BZ("end"),
    instructions.Load(0),
    instructions.Int(1),
    instructions.Add(),
    instructions.Store(0),
    instructions.B("loop"),
    instructions.Label("end"),
    instructions.Int(1),
    instructions.Return(),
]

ins_partitions = [
    (0, 5),
    (5, 7),
    (7, 12),
    (12, 16),
    (16, 20),
    (20, 24),
    (24, 25),
    (25, 26),
    (26, 27),
    (27, 30),
    (30, 32),
    (32, 37),
    (37, 42),
    (42, 45),
]

bbs_links = [
    (0, 1),
    (0, 2),
    (2, 3),
    (2, 7),
    (3, 4),
    (3, 8),
    (4, 5),
    (4, 9),
    (5, 6),
    (5, 10),
    (7, 8),
    (8, 9),
    (6, 11),
    (11, 12),
    (11, 13),
    (12, 11),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

CAN_UPDATE_LOOP_VULNERABLE_PATHS = [
    [bbs[0], bbs[2], bbs[3], bbs[4], bbs[5], bbs[6], bbs[11], bbs[13]],
]

can_update_tests = [
    (CAN_UPDATE, IsUpdatable, CAN_UPDATE_VULNERABLE_PATHS),
    (CAN_UPDATE_LOOP, IsUpdatable, CAN_UPDATE_LOOP_VULNERABLE_PATHS),
]

CAN_UPDATE_GROUP_INDEX_0 = """
#pragma version 6
txn GroupIndex
int 0
==
assert
int 0
gtxn 0 ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    int NoOp
    ==
    bnz handle_noop
    gtxn 0 OnCompletion
    int OptIn
    ==
    bnz handle_optin
    gtxn 0 OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    gtxn 0 OnCompletion
    int UpdateApplication
    ==
    bnz handle_updateapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_0_VULNERABLE_PATHS: List[List[int]] = []

CAN_UPDATE_GROUP_INDEX_1 = """
#pragma version 6
global GroupSize
int 1
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    int NoOp
    ==
    bnz handle_noop
    gtxn 0 OnCompletion
    int OptIn
    ==
    bnz handle_optin
    gtxn 0 OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    txn OnCompletion
    int UpdateApplication
    ==
    bnz handle_updateapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_1_VULNERABLE_PATHS: List[List[int]] = []

CAN_UPDATE_GROUP_INDEX_2 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 1 OnCompletion
    int NoOp
    ==
    bnz handle_noop
    gtxn 1 OnCompletion
    int OptIn
    ==
    bnz handle_optin
    gtxn 1 OnCompletion
    int CloseOut
    ==
    bz handle_closeout
    gtxn 0 OnCompletion
    int UpdateApplication
    ==
    bnz handle_updateapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS: List[List[int]] = [
    [0, 2, 3, 4, 9],
    [0, 2, 3, 8, 9],
    [0, 2, 7, 8, 9],
]

CAN_UPDATE_GROUP_INDEX_INTC_0 = """
#pragma version 6
intcblock 0 1 0x2 0x9301 0x3 0x4 0x5
txn GroupIndex
int 0
==
assert
int 0
gtxn 0 ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    intc_0
    ==
    bnz handle_noop
    gtxn 0 OnCompletion
    intc_1
    ==
    bnz handle_optin
    gtxn 0 OnCompletion
    intc_2
    ==
    bnz handle_closeout
    gtxn 0 OnCompletion
    intc 5
    ==
    bnz handle_updateapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_INTC_0_VULNERABLE_PATHS: List[List[int]] = []

CAN_UPDATE_GROUP_INDEX_2_X_0 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 1 OnCompletion
    int OptIn
    ==
    ||
    gtxn 1 OnCompletion
    int CloseOut
    ==
    ||                      // (Eq || Eq) || Eq)
    bnz success             // gtxn 1 is one of NoOp, OptIn, CloseOut. gtxn 0 can be UpdateApplication
    gtxn 1 OnCompletion
    int UpdateApplication
    ==
    gtxn 1 OnCompletion
    int DeleteApplication
    ==
    ||
    bnz handle_updateapp        // pattern: (Eq<> || Eq<>)
    int 1
    return
success:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_0: List[List[int]] = [
    [0, 2, 3, 4],
    [0, 2, 5],
]

CAN_UPDATE_GROUP_INDEX_2_X_1 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    int NoOp
    ==
    gtxn 0 OnCompletion
    int OptIn
    ==
    ||
    !
    !
    gtxn 0 OnCompletion
    int CloseOut
    ==
    ||
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 1 OnCompletion
    int OptIn
    !=
    !
    ||
    &&                              // gtxn 0 is NoOp or OptIn or CloseOut and gtxn 1 is NoOp or OptIn.
    bnz success                     // Pattern:    (!!(== || ==) || ==) && (== || !(!=))
    gtxn 1 OnCompletion
    int UpdateApplication
    ==
    gtxn 1 OnCompletion
    int DeleteApplication
    ==
    ||
    bnz handle_updateapp
    int 1
    return
success:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_1: List[List[int]] = [
    [0, 2, 3, 4],
]

CAN_UPDATE_GROUP_INDEX_2_X_2 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    int NoOp
    ==
    !
    gtxn 0 OnCompletion
    int OptIn
    ==
    !
    &&
    !                       // !(&&(!x, !y)) => x || y
    gtxn 0 OnCompletion
    int CloseOut
    ==
    ||
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 1 OnCompletion
    int OptIn
    !=
    !
    ||
    &&                          // Pattern: (!(&&(!x, !y)) || ==) && (== || !(!=))
    bnz success                 // gtxn 0 is NoOp or OptIn or CloseOut and gtxn 1 is NoOp or OptIn. max group size is 2
    gtxn 1 OnCompletion
    int UpdateApplication
    ==
    gtxn 1 OnCompletion
    int DeleteApplication
    ==
    ||
    bnz handle_updateapp
    int 1
    return
success:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_2: List[List[int]] = [
    [0, 2, 3, 4],
]

CAN_UPDATE_GROUP_INDEX_2_X_3 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    int NoOp
    ==
    !
    gtxn 0 OnCompletion
    int OptIn
    ==
    !
    &&
    !                       // !(!x && !y) => x || y
    gtxn 0 OnCompletion
    int CloseOut
    ==
    ||
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 1 OnCompletion
    int OptIn
    !=
    !
    ||
    &&
    bnz success                 // gtxn 0 is NoOp or OptIn or CloseOut and gtxn 1 is NoOp or OptIn.
    gtxn 1 OnCompletion
    int UpdateApplication
    !=
    gtxn 1 OnCompletion
    int DeleteApplication
    !=
    &&
    !                           // !(!x && !y) => x || y. True if gtxn 1 is UpdateApplication or DeleteApplication
    gtxn 0 OnCompletion
    int UpdateApplication
    ==
    gtxn 0 OnCompletion
    int DeleteApplication
    !=
    !
    ||                          // True if gtxn 0 is UpdateApplication or DeleteApplication.
    ||
    bnz handle_updateapp
    int 1
    return
success:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_3: List[List[int]] = []

CAN_UPDATE_GROUP_INDEX_2_X_4 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    int NoOp
    ==
    ||                      // depends on a unknown value
    gtxn 0 OnCompletion
    int OptIn
    ==
    ||
    gtxn 0 OnCompletion
    int CloseOut
    ==
    ||                          // gtxn 0 does not have to be NoOp, OptIn, CloseOut. 
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 1 OnCompletion
    int OptIn
    ==
    ||
    &&
    bnz success                 // gtxn 1 is NoOp or OptIn.
    gtxn 1 OnCompletion
    int UpdateApplication
    !=
    gtxn 1 OnCompletion
    int DeleteApplication
    !=
    &&
    !                           // !(!x && !y) => x || y. True if gtxn 1 is UpdateApplication or DeleteApplication
    gtxn 0 OnCompletion
    int UpdateApplication
    ==
    gtxn 0 OnCompletion
    int DeleteApplication
    !=
    !
    ||                          // True if gtxn 0 is UpdateApplication or DeleteApplication.
    ||
    bnz handle_updateapp
    int 1
    return
success:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_4: List[List[int]] = [
    [0, 2, 5],  # gtxn 0 can be UpdateApplication or DeleteApplication and reach "success:" block
]

CAN_UPDATE_GROUP_INDEX_2_X_5 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 0 OnCompletion
    int NoOp
    ==
    ||                      // depends on a independent value. GTXN_0_TransactionType is different from GTXN_1_TransactionType
    gtxn 0 OnCompletion
    int OptIn
    ==
    ||
    gtxn 0 OnCompletion
    int CloseOut
    ==
    ||                          // gtxn 0 does not have to be NoOp, OptIn, CloseOut. 
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 1 OnCompletion
    int OptIn
    ==
    ||
    &&
    bnz success                 // gtxn 1 is NoOp or OptIn.
    gtxn 1 OnCompletion
    int UpdateApplication
    !=
    gtxn 1 OnCompletion
    int DeleteApplication
    !=
    &&
    !                           // !(!x && !y) => x || y. True if gtxn 1 is UpdateApplication or DeleteApplication
    gtxn 0 OnCompletion
    int UpdateApplication
    ==
    gtxn 0 OnCompletion
    int DeleteApplication
    !=
    !
    ||                          // True if gtxn 0 is UpdateApplication or DeleteApplication.
    ||
    bnz handle_updateapp
    int 1
    return
success:
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_5: List[List[int]] = [
    [0, 2, 5],  # gtxn 0 can be UpdateApplication or DeleteApplication and reach "success:" block
]

CAN_UPDATE_GROUP_INDEX_2_X_6 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 0 OnCompletion
    int NoOp
    ==
    ||                      // depends on a independent value. GTXN_0_TransactionType is different from GTXN_1_TransactionType
    gtxn 0 OnCompletion
    int OptIn
    ==
    ||
    gtxn 0 OnCompletion
    int CloseOut
    ==
    ||                          // gtxn 0 does not have to be NoOp, OptIn, CloseOut. 
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 1 OnCompletion
    int OptIn
    ==
    ||
    &&
    bnz success                 // gtxn 1 is NoOp or OptIn.
    gtxn 1 OnCompletion
    int UpdateApplication
    !=
    gtxn 1 OnCompletion
    int DeleteApplication
    !=
    &&
    !                           // !(!x && !y) => x || y. True if gtxn 1 is UpdateApplication or DeleteApplication
    gtxn 0 OnCompletion
    int UpdateApplication
    ==
    gtxn 0 OnCompletion
    int DeleteApplication
    !=
    !
    ||                          // True if gtxn 0 is UpdateApplication or DeleteApplication.
    ||
    bnz handle_updateapp
    int 1
    return
success:
    gtxn 0 OnCompletion
    int UpdateApplication
    !=
    gtxn 0 OnCompletion
    int DeleteApplication
    !=
    &&
    assert
    int 1
    return
handle_updateapp:
    err
"""

CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_6: List[List[int]] = [
    # gtxn 0 can be UpdateApplication or DeleteApplication and reach "success:" block
    # success block checks for gtxn 0 UpdateApplication and DeleteApplication
]

CAN_UPDATE_GROUP_INDEX_2_X_7 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 0 OnCompletion
    int NoOp
    ==
    ||                      // depends on a independent value. GTXN_0_TransactionType is different from GTXN_1_TransactionType
    gtxn 0 OnCompletion
    int OptIn
    ==
    ||
    gtxn 0 OnCompletion
    int CloseOut
    ==
    ||                          // gtxn 0 does not have to be NoOp, OptIn, CloseOut. 
    gtxn 1 OnCompletion
    int NoOp
    ==
    gtxn 1 OnCompletion
    int OptIn
    ==
    ||
    &&
    bnz success                 // gtxn 1 is NoOp or OptIn.
    gtxn 1 OnCompletion
    int UpdateApplication
    !=
    gtxn 1 OnCompletion
    int DeleteApplication
    !=
    &&
    !                           // !(!x && !y) => x || y. True if gtxn 1 is UpdateApplication or DeleteApplication
    gtxn 0 OnCompletion
    int UpdateApplication
    ==
    gtxn 0 OnCompletion
    int DeleteApplication
    !=
    !
    ||                          // True if gtxn 0 is UpdateApplication or DeleteApplication.
    ||
    bnz handle_updateapp
    int 1
    return
success:
gtxn 0 OnCompletion
int UpdateApplication
!=
txn Fee
int 10000
<
&&                          // gtxn 0 Completion should not be UpdateApplication for this to be true
gtxn 0 OnCompletion
int UpdateApplication
!=
txn Note
byte "Note"
==
&&                          // gtxn 0 Completion should not be UpdateApplication for this to be true
||                          // r = ( x && a ) || ( x && b), x must be True for r to be True, x here is `Neq(gtxn 0 OnCompletion, UpdateApplication)`
assert
int 1
return
handle_updateapp:
err
"""

CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_7: List[List[int]] = [
    # gtxn 0 can be UpdateApplication or DeleteApplication and reach "success:" block
    # success block checks for gtxn 0 OnCompletion is not UpdateApplication
]

CAN_DELETE_GROUP_INDEX_2_VULNERABLE_PATHS_X_7: List[List[int]] = [
    # gtxn 0 can be UpdateApplication or DeleteApplication and reach "success:" block
    # success block checks for gtxn 0 OnCompletion is not UpdateApplication but not DeleteApplication
    [0, 2, 5],
]

new_can_update_tests: List[Tuple[str, Type[AbstractDetector], List[List[int]]]] = [
    (CAN_UPDATE_GROUP_INDEX_0, IsUpdatable, CAN_UPDATE_GROUP_INDEX_0_VULNERABLE_PATHS),
    (CAN_UPDATE_GROUP_INDEX_1, IsUpdatable, CAN_UPDATE_GROUP_INDEX_1_VULNERABLE_PATHS),
    (CAN_UPDATE_GROUP_INDEX_2, IsUpdatable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS),
    (CAN_UPDATE_GROUP_INDEX_INTC_0, IsUpdatable, CAN_UPDATE_GROUP_INDEX_INTC_0_VULNERABLE_PATHS),
    (
        CAN_UPDATE,
        CanCloseAccount,
        [],
    ),  # Applications are not vulnerable to CanCloseAccount and CanCloseAsset
    (CAN_UPDATE_LOOP, CanCloseAccount, []),
    (CAN_UPDATE_GROUP_INDEX_0, CanCloseAccount, []),
    (CAN_UPDATE_GROUP_INDEX_1, CanCloseAccount, []),
    (CAN_UPDATE_GROUP_INDEX_2, CanCloseAccount, []),
    (CAN_UPDATE_GROUP_INDEX_INTC_0, CanCloseAccount, []),
    (CAN_UPDATE, CanCloseAsset, []),
    (CAN_UPDATE_LOOP, CanCloseAsset, []),
    (CAN_UPDATE_GROUP_INDEX_0, CanCloseAsset, []),
    (CAN_UPDATE_GROUP_INDEX_1, CanCloseAsset, []),
    (CAN_UPDATE_GROUP_INDEX_2, CanCloseAsset, []),
    (CAN_UPDATE_GROUP_INDEX_INTC_0, CanCloseAsset, []),
    (CAN_UPDATE_GROUP_INDEX_2_X_0, IsUpdatable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_0),
    (CAN_UPDATE_GROUP_INDEX_2_X_0, IsDeletable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_0),
    (CAN_UPDATE_GROUP_INDEX_2_X_1, IsUpdatable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_1),
    (CAN_UPDATE_GROUP_INDEX_2_X_2, IsDeletable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_1),
    (CAN_UPDATE_GROUP_INDEX_2_X_2, IsUpdatable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_2),
    (CAN_UPDATE_GROUP_INDEX_2_X_2, IsDeletable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_2),
    (CAN_UPDATE_GROUP_INDEX_2_X_3, IsUpdatable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_3),
    (CAN_UPDATE_GROUP_INDEX_2_X_3, IsDeletable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_3),
    (CAN_UPDATE_GROUP_INDEX_2_X_4, IsUpdatable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_4),
    (CAN_UPDATE_GROUP_INDEX_2_X_4, IsDeletable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_4),
    (CAN_UPDATE_GROUP_INDEX_2_X_5, IsUpdatable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_5),
    (CAN_UPDATE_GROUP_INDEX_2_X_5, IsDeletable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_5),
    (CAN_UPDATE_GROUP_INDEX_2_X_6, IsUpdatable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_6),
    (CAN_UPDATE_GROUP_INDEX_2_X_6, IsDeletable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_6),
    (CAN_UPDATE_GROUP_INDEX_2_X_7, IsUpdatable, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_7),
    (CAN_UPDATE_GROUP_INDEX_2_X_7, IsDeletable, CAN_DELETE_GROUP_INDEX_2_VULNERABLE_PATHS_X_7),
]

with open("tests/detectors/detector_test_files/app_approval_689168853.teal", encoding="utf-8") as f:
    APPROVAL_PROGRAM_689168853 = f.read()

# Sender is not verified in any of the above IsUpdatable, IsDeletable tests.
new_anyone_can_update_tests: List[Tuple[str, Type[AbstractDetector], List[List[int]]]] = [
    (CAN_UPDATE_GROUP_INDEX_0, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_0_VULNERABLE_PATHS),
    (CAN_UPDATE_GROUP_INDEX_1, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_1_VULNERABLE_PATHS),
    (CAN_UPDATE_GROUP_INDEX_2, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS),
    (CAN_UPDATE_GROUP_INDEX_2_X_0, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_0),
    (CAN_UPDATE_GROUP_INDEX_2_X_0, AnyoneCanDelete, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_0),
    (CAN_UPDATE_GROUP_INDEX_2_X_1, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_1),
    (CAN_UPDATE_GROUP_INDEX_2_X_2, AnyoneCanDelete, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_1),
    (CAN_UPDATE_GROUP_INDEX_2_X_2, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_2),
    (CAN_UPDATE_GROUP_INDEX_2_X_2, AnyoneCanDelete, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_2),
    (CAN_UPDATE_GROUP_INDEX_2_X_3, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_3),
    (CAN_UPDATE_GROUP_INDEX_2_X_3, AnyoneCanDelete, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_3),
    (CAN_UPDATE_GROUP_INDEX_2_X_4, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_4),
    (CAN_UPDATE_GROUP_INDEX_2_X_4, AnyoneCanDelete, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_4),
    (CAN_UPDATE_GROUP_INDEX_2_X_5, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_5),
    (CAN_UPDATE_GROUP_INDEX_2_X_5, AnyoneCanDelete, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_5),
    (CAN_UPDATE_GROUP_INDEX_2_X_6, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_6),
    (CAN_UPDATE_GROUP_INDEX_2_X_6, AnyoneCanDelete, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_6),
    (CAN_UPDATE_GROUP_INDEX_2_X_7, AnyoneCanUpdate, CAN_UPDATE_GROUP_INDEX_2_VULNERABLE_PATHS_X_7),
    (CAN_UPDATE_GROUP_INDEX_2_X_7, AnyoneCanDelete, CAN_DELETE_GROUP_INDEX_2_VULNERABLE_PATHS_X_7),
    # example from benchmarks
    (APPROVAL_PROGRAM_689168853, AnyoneCanUpdate, []),
    (APPROVAL_PROGRAM_689168853, AnyoneCanDelete, []),
]


new_can_update_tests += new_anyone_can_update_tests



# pylint: disable=undefined-variable
# type: ignore[name-defined]
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import

barecalls = BareCallActions(
    no_op=OnCompleteAction(action=Approve(), call_config=CallConfig.CREATE),
    opt_in=OnCompleteAction(action=Approve(), call_config=CallConfig.ALL),
    close_out=OnCompleteAction(action=Approve(), call_config=CallConfig.CALL),
    update_application=OnCompleteAction(call_config=CallConfig.NEVER),
    delete_application=OnCompleteAction(call_config=CallConfig.NEVER),
)

router = Router(
    name="ARC4 application parsing test",
    bare_calls=barecalls,
    clear_state=Approve(),
)


@ABIReturnSubroutine
def identity(arg: abi.Uint64, *, output: abi.Uint64) -> Expr:
    return output.set(arg.get())


@router.method(  # type: ignore[misc]
    opt_in=CallConfig.ALL,
    close_out=CallConfig.CALL,
    update_application=CallConfig.NEVER,
    delete_application=CallConfig.NEVER,
)
def hello(name: abi.String, *, output: abi.String) -> Expr:
    return output.set(Concat(Bytes("Hello "), name.get()))


router.add_method_handler(identity, method_config=MethodConfig(no_op=CallConfig.CALL))


arc4_application_ap, _clear_program, _ = router.compile_program(version=6)



from typing import List, Tuple, Type

from tealer.teal.instructions import instructions
from tealer.teal.instructions import transaction_field
from tealer.detectors.abstract_detector import AbstractDetector
from tealer.detectors.all_detectors import IsDeletable, CanCloseAccount, CanCloseAsset

from tests.utils import construct_cfg


CAN_DELETE = """
#pragma version 2
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    txn OnCompletion
    int NoOp
    ==
    bnz handle_noop
    txn OnCompletion
    int OptIn
    ==
    bnz handle_optin
    txn OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    txn OnCompletion
    int UpdateApplication
    ==
    bnz handle_updateapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_updateapp:
err
"""

ins_list = [
    instructions.Pragma(2),
    instructions.Int(0),
    instructions.Txn(transaction_field.ApplicationID()),
    instructions.Eq(),
    instructions.BZ("not_creation"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("not_creation"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("NoOp"),
    instructions.Eq(),
    instructions.BNZ("handle_noop"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("OptIn"),
    instructions.Eq(),
    instructions.BNZ("handle_optin"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("CloseOut"),
    instructions.Eq(),
    instructions.BNZ("handle_closeout"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("UpdateApplication"),
    instructions.Eq(),
    instructions.BNZ("handle_updateapp"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("handle_noop"),
    instructions.Label("handle_optin"),
    instructions.Label("handle_closeout"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("handle_updateapp"),
    instructions.Err(),
]

ins_partitions = [
    (0, 5),
    (5, 7),
    (7, 12),
    (12, 16),
    (16, 20),
    (20, 24),
    (24, 26),
    (26, 27),
    (27, 28),
    (28, 31),
    (31, 33),
]

bbs_links = [
    (0, 1),
    (0, 2),
    (2, 3),
    (2, 7),
    (3, 4),
    (3, 8),
    (4, 5),
    (4, 9),
    (5, 6),
    (5, 10),
    (7, 8),
    (8, 9),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

CAN_DELETE_VULNERABLE_PATHS = [
    [bbs[0], bbs[2], bbs[3], bbs[4], bbs[5], bbs[6]],
]


CAN_DELETE_LOOP = """
#pragma version 4
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    txn OnCompletion
    int NoOp
    ==
    bnz handle_noop
    txn OnCompletion
    int OptIn
    ==
    bnz handle_optin
    txn OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    txn OnCompletion
    int UpdateApplication
    ==
    bnz handle_updateapp
    b loop
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_updateapp:
err
loop:
    load 0
    int 5
    <
    bz end
    load 0
    int 1
    +
    store 0
    b loop
end:
    int 1
    return
"""

ins_list = [
    instructions.Pragma(4),
    instructions.Int(0),
    instructions.Txn(transaction_field.ApplicationID()),
    instructions.Eq(),
    instructions.BZ("not_creation"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("not_creation"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("NoOp"),
    instructions.Eq(),
    instructions.BNZ("handle_noop"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("OptIn"),
    instructions.Eq(),
    instructions.BNZ("handle_optin"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("CloseOut"),
    instructions.Eq(),
    instructions.BNZ("handle_closeout"),
    instructions.Txn(transaction_field.OnCompletion()),
    instructions.Int("UpdateApplication"),
    instructions.Eq(),
    instructions.BNZ("handle_updateapp"),
    instructions.B("loop"),
    instructions.Label("handle_noop"),
    instructions.Label("handle_optin"),
    instructions.Label("handle_closeout"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("handle_updateapp"),
    instructions.Err(),
    instructions.Label("loop"),
    instructions.Load(0),
    instructions.Int(5),
    instructions.Less(),
    instructions.BZ("end"),
    instructions.Load(0),
    instructions.Int(1),
    instructions.Add(),
    instructions.Store(0),
    instructions.B("loop"),
    instructions.Label("end"),
    instructions.Int(1),
    instructions.Return(),
]

ins_partitions = [
    (0, 5),
    (5, 7),
    (7, 12),
    (12, 16),
    (16, 20),
    (20, 24),
    (24, 25),
    (25, 26),
    (26, 27),
    (27, 30),
    (30, 32),
    (32, 37),
    (37, 42),
    (42, 45),
]

bbs_links = [
    (0, 1),
    (0, 2),
    (2, 3),
    (2, 7),
    (3, 4),
    (3, 8),
    (4, 5),
    (4, 9),
    (5, 6),
    (5, 10),
    (7, 8),
    (8, 9),
    (6, 11),
    (11, 12),
    (11, 13),
    (12, 11),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

CAN_DELETE_LOOP_VULNERABLE_PATHS = [
    [bbs[0], bbs[2], bbs[3], bbs[4], bbs[5], bbs[6], bbs[11], bbs[13]],
]


can_delete_tests = [
    (CAN_DELETE, IsDeletable, CAN_DELETE_VULNERABLE_PATHS),
    (CAN_DELETE_LOOP, IsDeletable, CAN_DELETE_LOOP_VULNERABLE_PATHS),
]


CAN_DELETE_GROUP_INDEX_0 = """
#pragma version 6
txn GroupIndex
int 0
==
assert
int 0
gtxn 0 ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 0 OnCompletion
    int NoOp
    ==
    bnz handle_noop
    gtxn 0 OnCompletion
    int OptIn
    ==
    bnz handle_optin
    gtxn 0 OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    gtxn 0 OnCompletion
    int UpdateApplication
    ==
    bnz handle_updateapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_updateapp:
err
"""

CAN_DELETE_GROUP_INDEX_0_VULNERABLE_PATHS: List[List[int]] = [[0, 2, 3, 4, 5, 6]]

CAN_DELETE_GROUP_INDEX_1 = """
#pragma version 6
global GroupSize
int 1
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    txn OnCompletion
    int NoOp
    ==
    bnz handle_noop
    gtxn 0 OnCompletion
    int OptIn
    ==
    bnz handle_optin
    gtxn 0 OnCompletion
    int CloseOut
    ==
    bnz handle_closeout
    gtxn 0 OnCompletion
    int DeleteApplication
    ==
    bnz handle_deleteapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_deleteapp:
err
"""

CAN_DELETE_GROUP_INDEX_1_VULNERABLE_PATHS: List[List[int]] = []

CAN_DELETE_GROUP_INDEX_2 = """
#pragma version 6
global GroupSize
int 2
==
assert
int 0
txn ApplicationID
==
bz not_creation
int 1
return
not_creation:
    gtxn 1 OnCompletion
    int NoOp
    ==
    bnz handle_noop
    gtxn 1 OnCompletion
    int OptIn
    ==
    bnz handle_optin
    gtxn 1 OnCompletion
    int CloseOut
    ==
    bz handle_closeout
    gtxn 0 OnCompletion
    int DeleteApplication
    ==
    bnz handle_deleteapp
    int 1
    return
handle_noop:
handle_optin:
handle_closeout:
int 1
return
handle_deleteapp:
err
"""

CAN_DELETE_GROUP_INDEX_2_VULNERABLE_PATHS: List[List[int]] = [
    [0, 2, 3, 4, 9],
    [0, 2, 3, 8, 9],
    [0, 2, 7, 8, 9],
]

new_can_delete_tests: List[Tuple[str, Type[AbstractDetector], List[List[int]]]] = [
    (CAN_DELETE_GROUP_INDEX_0, IsDeletable, CAN_DELETE_GROUP_INDEX_0_VULNERABLE_PATHS),
    (CAN_DELETE_GROUP_INDEX_1, IsDeletable, CAN_DELETE_GROUP_INDEX_1_VULNERABLE_PATHS),
    (CAN_DELETE_GROUP_INDEX_2, IsDeletable, CAN_DELETE_GROUP_INDEX_2_VULNERABLE_PATHS),
    (CAN_DELETE, CanCloseAccount, []),
    (CAN_DELETE_LOOP, CanCloseAccount, []),
    (CAN_DELETE_GROUP_INDEX_0, CanCloseAccount, []),
    (CAN_DELETE_GROUP_INDEX_1, CanCloseAccount, []),
    (CAN_DELETE_GROUP_INDEX_2, CanCloseAccount, []),
    (CAN_DELETE, CanCloseAsset, []),
    (CAN_DELETE_LOOP, CanCloseAsset, []),
    (CAN_DELETE_GROUP_INDEX_0, CanCloseAsset, []),
    (CAN_DELETE_GROUP_INDEX_1, CanCloseAsset, []),
    (CAN_DELETE_GROUP_INDEX_2, CanCloseAsset, []),
]



"""Printer for exporting call-graph of the contract."""

import html
import os
from pathlib import Path
from typing import Dict, Set

from tealer.printers.abstract_printer import AbstractPrinter
from tealer.utils.output import ROOT_OUTPUT_DIRECTORY


class PrinterCallGraph(AbstractPrinter):  # pylint: disable=too-few-public-methods
    """Printer to export call-graph of the contract.

    This printer is supposed to work for contracts written in teal version 4 or greater.
    Call graph dot file will be saved as `call-graph.dot` in destination folder if given
    or else in the current directory. The entry point of the contract is
    treated as another function("__entry__") while constructing the call graph.

    """

    NAME = "call-graph"
    HELP = "Export the call graph of contract to a dot file"
    WIKI_URL = "https://github.com/crytic/tealer/wiki/Printer-documentation#call-graph"

    def _construct_call_graph(self) -> Dict[str, Set[str]]:
        """construct call graph for the contract.

        entry point is treated as a separate subroutine/function `__main__`.
        For each subroutine, label(subroutine name) represents the node and each callsub
        instruction in the subroutine corresponds to directed edge to the target subroutine.

        Returns:
            Dict[str, Set[str]]: dictionary representing the graph. The graph has edges from each
            subroutine in d[key] to key. The source node of the edge are values and destination node is the key.
            if d["S1"] = ["S2", "S3"] then graph has edges "S3 -> S1", "S2 -> S1".
        """

        graph: Dict[str, Set[str]] = {}
        for _, subroutine in self.teal.subroutines.items():
            graph[subroutine.name] = set(
                map(lambda bi: bi.subroutine.name, subroutine.caller_blocks)
            )

        # no need to handle __main__ subroutine cause it is program entry point and there won't be caller blocks.
        return graph

    def print(self) -> None:
        """Export call graph of the contract in dot format.

        Each node in the call graph corresponds to a subroutine and edge representing call from
        source subroutine to called subroutine. Entry point of the contract is treated as another
        function and represented with label `__entry__`. Call graph will be saved as `call-graph.dot`
        in the destination directory if given or else in the current directory.
        Subroutines are supported from teal version 4. so, for contracts written in lesser version, this
        function only prints the error message stating the same."""

        if self.teal.version < 4:
            print("subroutines are not supported in teal version 3 or less")
            return

        dot_output = "digraph g{\n"
        graph = self._construct_call_graph()

        # construct dot representation of the graph
        graph_edges = ""
        for destination_sub, source_subs in graph.items():
            destination_sub = html.escape(destination_sub, quote=True)
            dot_output += f"{destination_sub}[label={destination_sub}];\n"
            for source_sub in source_subs:
                source_sub = html.escape(source_sub, quote=True)
                graph_edges += f"{source_sub} -> {destination_sub};\n"
        dot_output += graph_edges
        dot_output += "}\n"

        # outputs a single file: set the dir to {ROOT_DIRECTORY}/{CONTRACT_NAME}
        dest = ROOT_OUTPUT_DIRECTORY / Path(self.teal.contract_name)
        os.makedirs(dest, exist_ok=True)

        filename = Path("call-graph.dot")
        filename = dest / filename

        print(f"\nExported call graph to {filename}")
        with open(filename, "w", encoding="utf-8") as f:
            f.write(dot_output)



"""Detector for finding execution paths missing DeleteApplication check."""

from typing import List, TYPE_CHECKING, Tuple

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)

from tealer.detectors.utils import (
    detect_missing_tx_field_validations_group,
    detect_missing_tx_field_validations_group_complete,
)
from tealer.utils.teal_enums import TealerTransactionType
from tealer.utils.output import ExecutionPaths

if TYPE_CHECKING:
    from tealer.utils.output import ListOutput
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.teal.context.block_transaction_context import BlockTransactionContext
    from tealer.teal.teal import Teal


class IsDeletable(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find execution paths missing DeleteApplication check.

    Stateful smart contracts(application) can be deleted in algorand. If the
    application transaction of type DeleteApplication is approved by the application,
    then the application will be deleted. Contracts can check the application
    transaction type using OnCompletion field.

    This detector tries to find execution paths that approve the application
    transaction("return 1") and doesn't check the OnCompletion field against
    DeleteApplication value. Execution paths that only execute if the application
    transaction is not DeleteApplication are excluded.
    """

    NAME = "is-deletable"
    DESCRIPTION = "Deletable Applications"
    TYPE = DetectorType.STATEFULL

    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = "https://github.com/crytic/tealer/wiki/Detector-Documentation#deletable-application"
    WIKI_TITLE = "Deletable Application"
    WIKI_DESCRIPTION = (
        "Application can be deleted by sending an `DeleteApplication` type application call. "
    )
    WIKI_EXPLOIT_SCENARIO = """
```py
@router.method(delete_application=CallConfig.CALL)
def delete_application() -> Expr:
    return Assert(Txn.sender() == Global.creator_address())
```

Eve steals application creator's private key and deletes the application. Application's assets are permanently lost.
"""

    WIKI_RECOMMENDATION = """
Do not approve `DeleteApplication` type application calls.
"""

    def detect(self) -> "ListOutput":
        """Detect execution paths with missing DeleteApplication check.

        Returns:
            ExecutionPaths instance containing the list of vulnerable execution
            paths along with name, check, impact, confidence and other detector
            information.
        """

        def checks_field(block_ctx: "BlockTransactionContext") -> bool:
            # return False if Txn Type can be DeleteApplication.
            # return True if Txn Type cannot be DeleteApplication.
            return not TealerTransactionType.ApplDeleteApplication in block_ctx.transaction_types

        # there should be a better to decide which function to call ??
        if self.tealer.output_group:
            # mypy complains if the value is returned directly. Uesd the second suggestion mentioned here:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#variance
            return list(
                detect_missing_tx_field_validations_group_complete(self.tealer, self, checks_field)
            )

        output: List[
            Tuple["Teal", List[List["BasicBlock"]]]
        ] = detect_missing_tx_field_validations_group(self.tealer, checks_field)
        detector_output: "ListOutput" = []
        for contract, vulnerable_paths in output:
            detector_output.append(ExecutionPaths(contract, self, vulnerable_paths))

        return detector_output



import logging
from typing import List, TYPE_CHECKING, Dict, Any, Type


from tealer.detectors.abstract_detector import AbstractDetector, DetectorClassification
from tealer.printers.abstract_printer import AbstractPrinter
from tealer.exceptions import TealerException


if TYPE_CHECKING:
    from tealer.teal.teal import Teal
    from tealer.utils.output import ListOutput
    from tealer.execution_context.transactions import GroupTransaction


# from slither: slither/slither.py
def _check_common_things(
    thing_name: str, cls: Any, base_cls: Any, instance_list: List[Any]
) -> None:
    """Check if the class is correct subclass and is unique.

    check if :cls: is subclass of :base_cls: and it's instance is not already
    present in :instance_list:. if either of the conditions fail, then an
    exception is raised.

    Args:
        thing_name: name of the feature(add-on) that will on the teal
            contract("detector" or "printer"). Used for error reporting.
        cls: Class representing the feature.
        base_cls: Base class for the feature.
        instance_list: List of objects already registered to run.

    Raises:
        TealerException: raises exception if the :cls: is not subclass of
            :base_cls: or if an object of :cls: is present in the
            :instance_list:.
    """

    if not issubclass(cls, base_cls) or cls is base_cls:
        raise TealerException(
            f"You can't register {cls.__name__} as a {thing_name}."
            f"You need to pass a class that inherits from {base_cls.__name__}"
        )

    if any(isinstance(obj, cls) for obj in instance_list):
        raise TealerException(f"You can't register {cls.__name__} twice.")


class Tealer:
    """Base class for the tool"""

    def __init__(
        self,
        contracts: Dict[str, "Teal"],
        groups: List["GroupTransaction"],
        output_group: bool = False,
    ):
        self._contracts = contracts
        self._groups = groups

        self._detectors: List[AbstractDetector] = []
        self._printers: List[AbstractPrinter] = []
        # debug parameter to decide on the type of output reported by the detectors. TODO: Remove later.
        self._output_group: bool = output_group

    @property
    def contracts(self) -> Dict[str, "Teal"]:
        return self._contracts

    @property
    def contracts_list(self) -> List["Teal"]:
        return list(self._contracts.values())

    @property
    def groups(self) -> List["GroupTransaction"]:
        return self._groups

    @property
    def output_group(self) -> bool:
        return self._output_group

    # from slither: Slither Class in slither/slither.py
    @property
    def detectors(self) -> List[AbstractDetector]:
        """return list of registered detectors.

        Returns:
            Returns the list of registered detectors.
        """
        return self._detectors

    # from slither: Slither Class in slither/slither.py
    @property
    def detectors_high(self) -> List[AbstractDetector]:
        """return list of registered detectors with impact high

        Returns:
            Returns the list of registered detectors with impact high.
        """
        return [d for d in self._detectors if d.IMPACT == DetectorClassification.HIGH]

    # from slither: Slither Class in slither/slither.py
    @property
    def detectors_medium(self) -> List[AbstractDetector]:
        """return list of registered detectors with impact medium

        Returns:
            Returns the list of registered detectors with impact medium.
        """
        return [d for d in self._detectors if d.IMPACT == DetectorClassification.MEDIUM]

    # from slither: Slither Class in slither/slither.py
    @property
    def detectors_low(self) -> List[AbstractDetector]:
        """return list of registered detectors with impact low

        Returns:
            Returns the list of registered detectors with impact low.
        """
        return [d for d in self._detectors if d.IMPACT == DetectorClassification.LOW]

    # from slither: Slither Class in slither/slither.py
    @property
    def detectors_informational(self) -> List[AbstractDetector]:
        """return list of registered detectors with impact informational

        Returns:
            Returns the list of registered detectors with impact informational.
        """
        return [d for d in self._detectors if d.IMPACT == DetectorClassification.INFORMATIONAL]

    # from slither: Slither Class in slither/slither.py
    @property
    def detectors_optimization(self) -> List[AbstractDetector]:
        """return list of registered detectors with impact optimization

        Returns:
            Returns the list of registered detectors with impact optimization.
        """
        return [d for d in self._detectors if d.IMPACT == DetectorClassification.OPTIMIZATION]

    # from slither: Slither Class in slither/slither.py
    @property
    def printers(self) -> List[AbstractPrinter]:
        """return list of registered printers.

        Returns:
            Returns the list of registered printers.
        """
        return self._printers

    # from slither: Slither Class in slither/slither.py
    def register_detector(self, detector_class: Type[AbstractDetector]) -> None:
        """Register detector to run on the contract.

        Args:
            detector_class: Class representing the detector.
        """

        _check_common_things("detector", detector_class, AbstractDetector, self._detectors)

        instance = detector_class(self)
        self._detectors.append(instance)

    # from slither: Slither Class in slither/slither.py
    def register_printer(self, printer_class: Type[AbstractPrinter]) -> None:
        """Register printer to run on the contract.

        Args:
            printer_class: Class representing the printer.
        """

        _check_common_things("printer", printer_class, AbstractPrinter, self._printers)
        # Run on the first contract for now. TODO: decide on how printers should behave in group context.
        instance = printer_class(list(self.contracts.values())[0])
        self._printers.append(instance)

    # from slither: Slither Class in slither/slither.py
    def run_detectors(self) -> List["ListOutput"]:
        """Run all the registered detectors.

        Returns:
            List of results, each result corresponds to the output
            of a single detector.
        """

        results = []
        logger = logging.getLogger("Tealer")
        for d in self._detectors:
            logger.debug(f'[+] Running detector "{d.NAME}"')
            results.append(d.detect())

        return results

    # from slither: Slither Class in slither/slither.py
    def run_printers(self) -> List:
        """Run all the registered printers.

        Returns:
            List of results, each result corresponds to the output
            of a single printer.
        """

        return [p.print() for p in self._printers]



"""Detect paths missing validations on sender field AND allows to delete the application."""

from typing import List, TYPE_CHECKING, Tuple

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.detectors.utils import (
    detect_missing_tx_field_validations_group,
    detect_missing_tx_field_validations_group_complete,
)
from tealer.utils.teal_enums import TealerTransactionType
from tealer.utils.output import ExecutionPaths


if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.utils.output import ListOutput
    from tealer.teal.context.block_transaction_context import BlockTransactionContext
    from tealer.teal.teal import Teal


class AnyoneCanDelete(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find execution paths missing validations on sender field AND allows to delete the application.

    Stateful smart contracts(application) can be deleted in Algorand. If the
    application transaction of type DeleteApplication is approved by the application,
    then the application will be deleted.

    This detector tries to find execution paths for which
        - OnCompletion can be DeleteApplication And
        - Transaction sender can be any address.
    """

    NAME = "unprotected-deletable"
    DESCRIPTION = "Unprotected Deletable Applications"
    TYPE = DetectorType.STATEFULL

    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = "https://github.com/crytic/tealer/wiki/Detector-Documentation#unprotected-deletable-application"
    WIKI_TITLE = "Unprotected Deletable Application"
    WIKI_DESCRIPTION = (
        "Application can be deleted by anyone. "
        "More at [building-secure-contracts/not-so-smart-contracts/algorand/access_controls]"
        "(https://github.com/crytic/building-secure-contracts/tree/master/not-so-smart-contracts/algorand/access_controls)."
    )
    WIKI_EXPLOIT_SCENARIO = """\
```py
@router.method(delete_application=CallConfig.CALL)
def delete_application() -> Expr:
    return Approve()
```

Eve calls `delete_application` method and deletes the application making its assets permanently inaccesible.
"""

    WIKI_RECOMMENDATION = """
- Avoid deletable applications.
- Add access controls to the vulnerable method.
"""

    def detect(self) -> "ListOutput":
        """Detect execution paths missing validations on sender field AND can delete the application .

        Returns:
            ExecutionPaths instance containing the list of vulnerable execution
            paths along with name, check, impact, confidence and other detector
            information.
        """

        def checks_field(block_ctx: "BlockTransactionContext") -> bool:
            # return False if Txn Type can be DeleteApplication AND sender can be any address.
            # return True otherwise
            return not (
                TealerTransactionType.ApplDeleteApplication in block_ctx.transaction_types
                and block_ctx.sender.any_addr
            )

        # there should be a better to decide which function to call ??
        if self.tealer.output_group:
            # mypy complains if the value is returned directly. Uesd the second suggestion mentioned here:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#variance
            return list(
                detect_missing_tx_field_validations_group_complete(self.tealer, self, checks_field)
            )

        output: List[
            Tuple["Teal", List[List["BasicBlock"]]]
        ] = detect_missing_tx_field_validations_group(self.tealer, checks_field)
        detector_output: "ListOutput" = []
        for contract, vulnerable_paths in output:
            detector_output.append(ExecutionPaths(contract, self, vulnerable_paths))

        return detector_output



# pylint: disable=undefined-variable
# type: ignore[name-defined]
from pathlib import Path
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import


def lsig2_case_7() -> Expr:
    return Return(Int(1))


def lsig2_case_8() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.rekey_to() == Global.zero_address(),
                    Txn.fee() < Int(10000),
                    Txn.close_remainder_to() == Global.zero_address(),
                    Txn.asset_close_to() == Global.zero_address(),
                    Txn.on_completion() != OnComplete.UpdateApplication,
                    Txn.on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_9() -> Expr:
    return Return(Int(1))


def lsig2_case_10() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[0].rekey_to() == Global.zero_address(),
                    Gtxn[1].rekey_to() == Global.zero_address(),
                    Gtxn[Int(0)].fee() < Int(100000),
                    Gtxn[1].fee() < Int(1000),
                    Gtxn[Int(0)].close_remainder_to() == Global.zero_address(),
                    Gtxn[1].close_remainder_to() == Global.zero_address(),
                    Gtxn[Int(0)].asset_close_to() == Global.zero_address(),
                    Gtxn[1].asset_close_to() == Global.zero_address(),
                    Gtxn[0].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[0].on_completion() != OnComplete.DeleteApplication,
                    Gtxn[Int(1)].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[Int(1)].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_11() -> Expr:
    return lsig2_case_10()


def lsig2_case_12() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[1].rekey_to() == Global.zero_address(),
                    Gtxn[1].fee() < Int(10000),
                    Gtxn[1].close_remainder_to() == Global.zero_address(),
                    Gtxn[1].asset_close_to() == Global.zero_address(),
                )
            ),
            Assert(
                And(
                    Gtxn[1].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[1].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_13() -> Expr:
    return Return(Int(1))


def lsig2_case_14() -> Expr:
    return Return(Int(1))


def lsig2_case_15() -> Expr:
    return Return(Int(1))


def lsig2_case_16() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.rekey_to() == Global.zero_address(),
                    Txn.fee() < Int(10000),
                    Txn.close_remainder_to() == Global.zero_address(),
                    Txn.asset_close_to() == Global.zero_address(),
                    Txn.on_completion() != OnComplete.UpdateApplication,
                    Txn.on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_17() -> Expr:
    return Return(Int(1))


def lsig2_case_18() -> Expr:
    return Return(Int(1))


def lsig2_case_19() -> Expr:
    return Seq(
        [
            Assert(Gtxn[Txn.group_index() - Int(4)].rekey_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() - Int(4)].fee() < Int(10000)),
            Assert(Gtxn[Txn.group_index() - Int(4)].close_remainder_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() - Int(4)].asset_close_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[Txn.group_index() - Int(4)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() - Int(4)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_20() -> Expr:
    return Seq(
        [
            Assert(Gtxn[Txn.group_index() + Int(5)].rekey_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(5)].fee() < Int(10000)),
            Assert(Gtxn[Txn.group_index() + Int(5)].close_remainder_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(5)].asset_close_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[Txn.group_index() + Int(5)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(5)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_21() -> Expr:
    return Return(Int(1))


def lsig2_case_22() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[Txn.group_index() - Int(1)].rekey_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() - Int(1)].fee() < Int(10000),
                    Gtxn[Txn.group_index() - Int(1)].close_remainder_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() - Int(1)].asset_close_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() - Int(1)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() - Int(1)].on_completion()
                    != OnComplete.DeleteApplication,
                    Txn.rekey_to() == Global.zero_address(),
                    Txn.fee() < Int(10000),
                    Txn.close_remainder_to() == Global.zero_address(),
                    Txn.asset_close_to() == Global.zero_address(),
                    Txn.on_completion() != OnComplete.UpdateApplication,
                    Txn.on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_23() -> Expr:
    return lsig2_case_22()


def lsig2_case_24() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.rekey_to() == Global.zero_address(),
                    Txn.fee() < Int(10000),
                    Txn.close_remainder_to() == Global.zero_address(),
                    Txn.asset_close_to() == Global.zero_address(),
                    Txn.on_completion() != OnComplete.UpdateApplication,
                    Txn.on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_25() -> Expr:
    return Return(Int(1))


def lsig2_case_26() -> Expr:
    return Return(Int(1))


def lsig2_case_27() -> Expr:
    return Return(Int(1))


def lsig2_case_28() -> Expr:
    return Seq(
        [
            Assert(Gtxn[Txn.group_index() - Int(4)].rekey_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() - Int(4)].fee() < Int(10000)),
            Assert(Gtxn[Txn.group_index() - Int(4)].close_remainder_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() - Int(4)].asset_close_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[Txn.group_index() - Int(4)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() - Int(4)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_29() -> Expr:
    return Return(Int(1))


def lsig2_case_30() -> Expr:
    return Return(Int(1))


def lsig2_case_31() -> Expr:
    return Seq(
        [
            Assert(Gtxn[Txn.group_index() - Int(4)].rekey_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() - Int(4)].fee() < Int(10000)),
            Assert(Gtxn[Txn.group_index() - Int(4)].close_remainder_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() - Int(4)].asset_close_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[Txn.group_index() - Int(4)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() - Int(4)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_32() -> Expr:
    return Seq(
        [
            Assert(Gtxn[Txn.group_index() - Int(4)].rekey_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() - Int(4)].fee() < Int(10000)),
            Assert(Gtxn[Txn.group_index() - Int(4)].close_remainder_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() - Int(4)].asset_close_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[Txn.group_index() - Int(4)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() - Int(4)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_33() -> Expr:
    return Seq(
        [
            Assert(Gtxn[Txn.group_index() + Int(5)].rekey_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(5)].fee() < Int(10000)),
            Assert(Gtxn[Txn.group_index() + Int(5)].close_remainder_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(5)].asset_close_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[Txn.group_index() + Int(5)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(5)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def lsig2_case_34() -> Expr:
    return Return(Int(1))


def lsig2_case_35() -> Expr:
    return Return(Int(1))


def logic_sig() -> Expr:
    test_case = Btoi(Arg(0))
    return Seq(
        [
            Cond(
                # [test_case == Int(1), lsig2_case_1()],
                # [test_case == Int(2), lsig2_case_2()],
                # [test_case == Int(3), lsig2_case_3()],
                # [test_case == Int(4), lsig2_case_4()],
                # [test_case == Int(5), lsig2_case_5()],
                # [test_case == Int(6), lsig2_case_6()],
                [test_case == Int(7), lsig2_case_7()],
                [test_case == Int(8), lsig2_case_8()],
                [test_case == Int(9), lsig2_case_9()],
                [test_case == Int(10), lsig2_case_10()],
                [test_case == Int(11), lsig2_case_11()],
                [test_case == Int(12), lsig2_case_12()],
                [test_case == Int(13), lsig2_case_13()],
                [test_case == Int(14), lsig2_case_14()],
                [test_case == Int(15), lsig2_case_15()],
                [test_case == Int(16), lsig2_case_16()],
                [test_case == Int(17), lsig2_case_17()],
                [test_case == Int(18), lsig2_case_18()],
                [test_case == Int(19), lsig2_case_19()],
                [test_case == Int(20), lsig2_case_20()],
                [test_case == Int(21), lsig2_case_21()],
                [test_case == Int(22), lsig2_case_22()],
                [test_case == Int(23), lsig2_case_23()],
                [test_case == Int(24), lsig2_case_24()],
                [test_case == Int(25), lsig2_case_25()],
                [test_case == Int(26), lsig2_case_26()],
                [test_case == Int(27), lsig2_case_27()],
                [test_case == Int(28), lsig2_case_28()],
                [test_case == Int(29), lsig2_case_29()],
                [test_case == Int(30), lsig2_case_30()],
                [test_case == Int(31), lsig2_case_31()],
                [test_case == Int(32), lsig2_case_32()],
                [test_case == Int(33), lsig2_case_33()],
                [test_case == Int(34), lsig2_case_34()],
                [test_case == Int(35), lsig2_case_35()],
            )
        ]
    )


def compile_program(output_file: Path) -> None:
    teal = compileTeal(logic_sig(), mode=Mode.Signature, version=7)
    with open(output_file, "w", encoding="utf-8") as f:
        f.write(teal)


if __name__ == "__main__":
    compile_program(Path("logicsig_2.teal"))



from typing import List

from tealer.teal.instructions import instructions, transaction_field
from tealer.detectors.all_detectors import MissingFeeCheck
from tealer.teal import global_field

from tests.utils import construct_cfg


MISSING_FEE_CHECK = """
#pragma version 2
txn Receiver
addr 6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM
==
bz wrongreceiver 
txn RekeyTo
global ZeroAddress
==
bz rekeying
txn CloseRemainderTo
global ZeroAddress
==
bz closing_account
txn AssetCloseTo
global ZeroAddress
==
bz closing_asset
global GroupSize
int 1
==
bz unexpected_group_size
int 1
return
wrongreceiver:
rekeying:
closing_account:
closing_asset:
unexpected_group_size:
    err
"""

ins_list = [
    instructions.Pragma(2),
    instructions.Txn(transaction_field.Receiver()),
    instructions.Addr("6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM"),
    instructions.Eq(),
    instructions.BZ("wrongreceiver"),
    instructions.Txn(transaction_field.RekeyTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("rekeying"),
    instructions.Txn(transaction_field.CloseRemainderTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_account"),
    instructions.Txn(transaction_field.AssetCloseTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_asset"),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(1),
    instructions.Eq(),
    instructions.BZ("unexpected_group_size"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("wrongreceiver"),
    instructions.Label("rekeying"),
    instructions.Label("closing_account"),
    instructions.Label("closing_asset"),
    instructions.Label("unexpected_group_size"),
    instructions.Err(),
]

ins_partitions = [
    (0, 5),
    (5, 9),
    (9, 13),
    (13, 17),
    (17, 21),
    (21, 23),
    (23, 24),
    (24, 25),
    (25, 26),
    (26, 27),
    (27, 29),
]

bbs_links = [
    (0, 1),
    (0, 6),
    (1, 2),
    (1, 7),
    (2, 3),
    (2, 8),
    (3, 4),
    (3, 9),
    (4, 5),
    (4, 10),
    (6, 7),
    (7, 8),
    (8, 9),
    (9, 10),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

MISSING_FEE_CHECK_VULNERABLE_PATHS = [
    [bbs[0], bbs[1], bbs[2], bbs[3], bbs[4], bbs[5]],
]


MISSING_FEE_CHECK_LOOP = """
#pragma version 4
txn Receiver
addr 6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM
==
bz wrongreceiver 
txn RekeyTo
global ZeroAddress
==
bz rekeying
txn CloseRemainderTo
global ZeroAddress
==
bz closing_account
txn AssetCloseTo
global ZeroAddress
==
bz closing_asset
global GroupSize
int 1
==
bz unexpected_group_size
int 0
b loop
wrongreceiver:
rekeying:
closing_account:
closing_asset:
unexpected_group_size:
    err
loop:
    dup
    int 5
    <
    bz end
    int 1
    +
    b loop
end:
    int 1
    return
"""

ins_list = [
    instructions.Pragma(4),
    instructions.Txn(transaction_field.Receiver()),
    instructions.Addr("6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM"),
    instructions.Eq(),
    instructions.BZ("wrongreceiver"),
    instructions.Txn(transaction_field.RekeyTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("rekeying"),
    instructions.Txn(transaction_field.CloseRemainderTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_account"),
    instructions.Txn(transaction_field.AssetCloseTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_asset"),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(1),
    instructions.Eq(),
    instructions.BZ("unexpected_group_size"),
    instructions.Int(0),
    instructions.B("loop"),
    instructions.Label("wrongreceiver"),
    instructions.Label("rekeying"),
    instructions.Label("closing_account"),
    instructions.Label("closing_asset"),
    instructions.Label("unexpected_group_size"),
    instructions.Err(),
    instructions.Label("loop"),
    instructions.Dup(),
    instructions.Int(5),
    instructions.Less(),
    instructions.BZ("end"),
    instructions.Int(1),
    instructions.Add(),
    instructions.B("loop"),
    instructions.Label("end"),
    instructions.Int(1),
    instructions.Return(),
]

ins_partitions = [
    (0, 5),
    (5, 9),
    (9, 13),
    (13, 17),
    (17, 21),
    (21, 23),
    (23, 24),
    (24, 25),
    (25, 26),
    (26, 27),
    (27, 29),
    (29, 34),
    (34, 37),
    (37, 40),
]

bbs_links = [
    (0, 1),
    (0, 6),
    (1, 2),
    (1, 7),
    (2, 3),
    (2, 8),
    (3, 4),
    (3, 9),
    (4, 5),
    (4, 10),
    (5, 11),
    (11, 12),
    (11, 13),
    (12, 11),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

MISSING_FEE_CHECK_LOOP_VULNERABLE_PATHS = [
    [bbs[0], bbs[1], bbs[2], bbs[3], bbs[4], bbs[5], bbs[11], bbs[13]]
]


missing_fee_check_tests = [
    (MISSING_FEE_CHECK, MissingFeeCheck, MISSING_FEE_CHECK_VULNERABLE_PATHS),
    (MISSING_FEE_CHECK_LOOP, MissingFeeCheck, MISSING_FEE_CHECK_LOOP_VULNERABLE_PATHS),
]


CHECKS_IN_ONE_BRANCH = """


"""

BASIC_1 = """
#pragma version 6
txn Fee
int 1000
>=                          // Fee >= 1000
assert
txn GroupIndex
int 0
==
bz fail_txn

callsub perform_validations

process_txn:
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    txn Fee
    int 10000
    >=                          // Fee >= 10000; Fee can be any value greater than 10000
    bz fail_txn
    
    retsub
"""

BASIC_1_VULNERABLE_PATHS: List[List[int]] = [[0, 1, 4, 5, 2]]

BASIC_2 = """
#pragma version 6
txn Fee
int 1000
>=                          // Fee >= 1000
assert
txn GroupIndex
int 0
==
bz fail_txn

callsub perform_validations

process_txn:
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    txn Fee
    int 10000
    <=                          // Fee <= 10000
    bz fail_txn

    retsub
"""

BASIC_2_VULNERABLE_PATHS: List[List[int]] = []


UNKNOWN_FEE = """
#pragma version 6
txn Fee
global MinTxnFee
<=
assert
txn GroupIndex
int 0
==
bz fail_txn

callsub perform_validations

process_txn:
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    int 1
    assert
    retsub
"""

UNKNOWN_FEE_VULNERABLE_PATHS: List[List[int]] = []

BASIC_3 = """
#pragma version 6

callsub perform_validations

process_txn:
    int 1
    return

perform_validations:
    txn GroupIndex
    int 5
    ==
    gtxn 5 Fee
    global MinTxnFee
    int 5
    *
    ==
    &&
    assert
    retsub
"""

BASIC_3_VULNERABLE_PATHS: List[List[int]] = []


LARGE_FEE = """
#pragma version 6       // 0
callsub validate_fee

callsub perform_validations     // 1

process_txn:                    // 2
    int 1
    return

fail_txn:                       // 3
    int 0
    return

perform_validations:            // 4
    int 1
    assert
    retsub

validate_fee:                   // 5
    txn Fee
    int 10000000000             // Fee is much larger than maximum transaction cost
    >
    bnz fail_txn

    retsub                      // 6
"""


LARGE_FEE_VULNERABLE_PATHS: List[List[int]] = [
    [0, 5, 6, 1, 4, 2],
]

MULTIPLE_GROUP_SIZES_IN_SUBROUTINES = """
#pragma version 6
txn Fee
int 10000
>=
bnz fail_txn
callsub validate_fee
callsub perform_validations

process_txn:
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    int 1
    assert
    retsub

validate_fee:
    global GroupSize
    int 2
    ==
    bnz group_size_2

    global GroupSize
    int 1
    ==
    gtxn 0 Fee
    int 10000
    <=
    &&
    assert
    retsub

    group_size_2:
        gtxn 0 Fee
        int 1000
        <
        assert
        gtxn 1 Fee
        int 1000
        <
        bz fail_txn

    retsub
"""

MULTIPLE_GROUP_SIZES_VULNERABLE_PATHS: List[List[int]] = []


CHECKS_IN_MULTIPLE_SUBROUTINES = """
#pragma version 6
global GroupSize
int 2
==
bnz group_size_2

global GroupSize
int 1
==
assert
callsub validate_fee_group_size_1
b process_txn

group_size_2:
    callsub validate_fee_group_size_2
    b process_txn

process_txn:
    callsub perform_validations
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    int 1
    assert
    retsub

validate_fee_group_size_1:
    gtxn 0 Fee
    int 10000
    <
    assert
    retsub

validate_fee_group_size_2:
    gtxn 0 Fee
    int 1000
    <=
    assert
    gtxn 1 Fee
    int 1000
    <=
    bz fail_txn
    retsub
"""

CHECKS_IN_MULTIPLE_SUBROUTINES_VULNERABLE_PATHS: List[List[int]] = []


CHECKS_IN_MULTIPLE_SUBROUTINES_2 = """
#pragma version 6
global GroupSize
int 2
==
bnz group_size_2

global GroupSize
int 1
==
assert
callsub validate_fee_group_size_1
b process_txn

group_size_2:
    callsub validate_fee_group_size_2
    b process_txn

process_txn:
    callsub perform_validations
    int 1
    return

fail_txn:
    int 0
    return

perform_validations:
    int 1
    assert
    retsub

validate_fee_group_size_1:
    gtxn 0 Fee
    int 10000
    <
    assert
    retsub

validate_fee_group_size_2:
    gtxn 0 Fee
    global MinTxnFee
    int 3
    *
    <=                  // detection of this pattern is possible because of StackValue emulation/analysis
    assert
    gtxn 1 Fee
    int 1000
    int 4
    *
    <=
    bz fail_txn
    retsub
"""


CHECKS_IN_MULTIPLE_SUBROUTINES_2_VULNERABLE_PATHS: List[List[int]] = []


CHECK_IN_ONE_PATH = """
#pragma version 6
txn RekeyTo
global ZeroAddress
==
global GroupSize
int 1
==
&&
txn Fee
int 0
==
&&
bnz process_txn

fail:
    err

process_txn:
    callsub perform_validations
    int 1
    return

perform_validations:
    int 1
    assert
    retsub
"""

CHECK_IN_ONE_PATH_VULNERABLE_PATHS: List[List[int]] = []

# The commented out test contracts contain a basic block which is part of __main__ CFG and a subroutine.
# TODO: Handle that case
new_missing_fee_tests = [
    # (BASIC_1, MissingFeeCheck, BASIC_1_VULNERABLE_PATHS),
    # (BASIC_2, MissingFeeCheck, BASIC_2_VULNERABLE_PATHS),
    (UNKNOWN_FEE, MissingFeeCheck, UNKNOWN_FEE_VULNERABLE_PATHS),
    (BASIC_3, MissingFeeCheck, BASIC_3_VULNERABLE_PATHS),
    (LARGE_FEE, MissingFeeCheck, LARGE_FEE_VULNERABLE_PATHS),
    # (MULTIPLE_GROUP_SIZES_IN_SUBROUTINES, MissingFeeCheck, MULTIPLE_GROUP_SIZES_VULNERABLE_PATHS),
    (
        CHECKS_IN_MULTIPLE_SUBROUTINES,
        MissingFeeCheck,
        CHECKS_IN_MULTIPLE_SUBROUTINES_VULNERABLE_PATHS,
    ),
    (
        CHECKS_IN_MULTIPLE_SUBROUTINES_2,
        MissingFeeCheck,
        CHECKS_IN_MULTIPLE_SUBROUTINES_2_VULNERABLE_PATHS,
    ),
    (CHECK_IN_ONE_PATH, MissingFeeCheck, CHECK_IN_ONE_PATH_VULNERABLE_PATHS),
]



from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="tealer",
    description="Teal analyzer.",
    url="https://github.com/crytic/tealer",
    author="Trail of Bits",
    version="0.1.0",
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "prettytable>=0.7.2",
        "py-algorand-sdk",
        "pycryptodomex",
        "requests",
        "pyyaml",
    ],
    extras_require={"dev": ["pylint==2.13.4", "black==22.3.0", "mypy==0.942", "pytest-cov"]},
    license="AGPL-3.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={"console_scripts": ["tealer = tealer.__main__:main"]},
)



from typing import List

from tealer.teal.instructions import instructions, transaction_field
from tealer.detectors.all_detectors import MissingGroupSize
from tealer.teal import global_field
from tealer.teal.basic_blocks import BasicBlock

from tests.utils import construct_cfg


MISSING_GROUP_SIZE = """
#pragma version 2
txn Receiver
addr 6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM
==
bz wrongreceiver
txn Fee
int 10000
<
bz highfee
txn RekeyTo
global ZeroAddress
==
bz rekeying
txn CloseRemainderTo
global ZeroAddress
==
bz closing_account
txn AssetCloseTo
global ZeroAddress
==
bz closing_asset
int 1
return
wrongreceiver:
highfee:
rekeying:
closing_account:
closing_asset:
    err
"""

ins_list = [
    instructions.Pragma(2),
    instructions.Txn(transaction_field.Receiver()),
    instructions.Addr("6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM"),
    instructions.Eq(),
    instructions.BZ("wrongreceiver"),
    instructions.Txn(transaction_field.Fee()),
    instructions.Int(10000),
    instructions.Less(),
    instructions.BZ("highfee"),
    instructions.Txn(transaction_field.RekeyTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("rekeying"),
    instructions.Txn(transaction_field.CloseRemainderTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_account"),
    instructions.Txn(transaction_field.AssetCloseTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_asset"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("wrongreceiver"),
    instructions.Label("highfee"),
    instructions.Label("rekeying"),
    instructions.Label("closing_account"),
    instructions.Label("closing_asset"),
    instructions.Err(),
]

ins_partitions = [
    (0, 5),
    (5, 9),
    (9, 13),
    (13, 17),
    (17, 21),
    (21, 23),
    (23, 24),
    (24, 25),
    (25, 26),
    (26, 27),
    (27, 29),
]

bbs_links = [
    (0, 1),
    (0, 6),
    (1, 2),
    (1, 7),
    (2, 3),
    (2, 8),
    (3, 4),
    (3, 9),
    (4, 5),
    (4, 10),
    (6, 7),
    (7, 8),
    (8, 9),
    (9, 10),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

MISSING_GROUP_SIZE_VULNERABLE_PATHS: List[List[BasicBlock]] = []


MISSING_GROUP_SIZE_LOOP = """
#pragma version 4
txn Receiver
addr 6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM
==
bz wrongreceiver 
txn Fee
int 10000
<
bz highfee
txn RekeyTo
global ZeroAddress
==
bz rekeying
txn CloseRemainderTo
global ZeroAddress
==
bz closing_account
txn AssetCloseTo
global ZeroAddress
==
bz closing_asset
int 0
b loop
wrongreceiver:
highfee:
rekeying:
closing_account:
closing_asset:
    err
loop:
    dup
    int 5
    <
    bz end
    int 1
    +
    b loop
end:
    int 1
    return
"""

ins_list = [
    instructions.Pragma(4),
    instructions.Txn(transaction_field.Receiver()),
    instructions.Addr("6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM"),
    instructions.Eq(),
    instructions.BZ("wrongreceiver"),
    instructions.Txn(transaction_field.Fee()),
    instructions.Int(10000),
    instructions.Less(),
    instructions.BZ("highfee"),
    instructions.Txn(transaction_field.RekeyTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("rekeying"),
    instructions.Txn(transaction_field.CloseRemainderTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_account"),
    instructions.Txn(transaction_field.AssetCloseTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_asset"),
    instructions.Int(0),
    instructions.B("loop"),
    instructions.Label("wrongreceiver"),
    instructions.Label("highfee"),
    instructions.Label("rekeying"),
    instructions.Label("closing_account"),
    instructions.Label("closing_asset"),
    instructions.Err(),
    instructions.Label("loop"),
    instructions.Dup(),
    instructions.Int(5),
    instructions.Less(),
    instructions.BZ("end"),
    instructions.Int(1),
    instructions.Add(),
    instructions.B("loop"),
    instructions.Label("end"),
    instructions.Int(1),
    instructions.Return(),
]

ins_partitions = [
    (0, 5),
    (5, 9),
    (9, 13),
    (13, 17),
    (17, 21),
    (21, 23),
    (23, 24),
    (24, 25),
    (25, 26),
    (26, 27),
    (27, 29),
    (29, 34),
    (34, 37),
    (37, 40),
]

bbs_links = [
    (0, 1),
    (0, 6),
    (1, 2),
    (1, 7),
    (2, 3),
    (2, 8),
    (3, 4),
    (3, 9),
    (4, 5),
    (4, 10),
    (6, 7),
    (7, 8),
    (8, 9),
    (9, 10),
    (5, 11),
    (11, 12),
    (11, 13),
    (12, 11),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

MISSING_GROUP_SIZE_LOOP_VULNERABLE_PATHS: List[List[BasicBlock]] = []


missing_group_size_tests = [
    (MISSING_GROUP_SIZE, MissingGroupSize, MISSING_GROUP_SIZE_VULNERABLE_PATHS),
    (MISSING_GROUP_SIZE_LOOP, MissingGroupSize, MISSING_GROUP_SIZE_LOOP_VULNERABLE_PATHS),
]



"""Defines a comparable enum class.

This module implements a comparable enum class which extends
``Enum`` class with capabilities for comparing two symbolic names
of defined in the same enumeration.

Classes:
    ComparableEnum: Enum subclass with additional comparison capability
        i.e enumerations defined with ComparableEnum can be compared.
"""

from enum import Enum

from typing import Any


# from slither: slither/utils/comparable_enum.py
class ComparableEnum(Enum):
    """Class to represent comparable enum types."""

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ComparableEnum):
            return self.value == other.value
        return False

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, ComparableEnum):
            return self.value != other.value
        return False

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, ComparableEnum):
            return self.value < other.value
        return False

    def __repr__(self) -> str:
        return f"{self.value}"

    def __hash__(self) -> int:
        return hash(self.value)



import tempfile
from pathlib import Path

from tealer.teal.instructions.instructions import Eq
from tealer.teal.parse_teal import parse_teal
from tealer.utils.regex.regex import parse_regex, match_regex, update_config
from tealer.utils.output import CFGDotConfig, full_cfg_to_dot


def test_parse_regex() -> None:
    txt = """
    some_label => 
   gtxn 0 Amount
   int 13337
   =="""

    regex = parse_regex(txt)
    assert regex.label == "some_label"
    assert len(regex.instructions) == 3
    assert isinstance(regex.instructions[-1], Eq)


def test_match() -> None:

    with open("./tests/regex/vote_approval.teal", "r", encoding="utf8") as f_codebase:
        teal = parse_teal(f_codebase.read())

    with open("./tests/regex/regex.txt", "r", encoding="utf8") as f_regex:
        regex = parse_regex(f_regex.read())

    matches, covered = match_regex(teal, regex)

    assert len(matches) == 1

    assert len(covered) == 21

    # Ensure the CFG generation does not crash
    config = CFGDotConfig()
    update_config(config, matches, covered)

    with tempfile.NamedTemporaryFile() as fp:
        full_cfg_to_dot(teal, config, Path(fp.name))



# pylint: skip-file
# pylint: disable=undefined-variable
# type: ignore[name-defined]

"""
Template for pyteal test:

File containing the pyteal code
```
# pylint: disable=undefined-variable
# type: ignore[name-defined]
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import

def approval_program() -> Expr:
    return Return(Int(1))

def clear_program() -> Expr:
    return Return(Int(1))

approval_program_str = compileTeal(
    approval_program(),
    mode=Mode.Application,
    version=7,
    optimize=OptimizeOptions(scratch_slots=False),
)

_clear_program = compileTeal(
    clear_program(), mode=Mode.Application, version=7, optimize=OptimizeOptions(scratch_slots=False)
)
```

Test file that imports the pyteal code
```
# type: ignore[unreachable]
import sys
import pytest

from tealer.teal.instructions import instructions
from tealer.teal.parse_teal import parse_teal

if not sys.version_info >= (3, 10):
    pytest.skip(reason="PyTeal based tests require python >= 3.10", allow_module_level=True)

# pylint: disable=wrong-import-position
# Place import statements after the version check
from tests.pyteal_parsing.normal_application import normal_application_approval_program
```
"""

from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import


@Subroutine(TealType.none)  # type: ignore[misc]
def foo_bar(n: Expr) -> Expr:
    return While(n < Int(10)).Do(
        If(n < Int(5))
        .Then(
            Log(Bytes("Foo")),
        )
        .Else(
            Log(Bytes("Bar")),
        )
    )


@Subroutine(TealType.none)  # type: ignore[misc]
def break_continue(n: ScratchVar) -> Expr:
    i = ScratchVar()
    return While(Int(1)).Do(
        If(n.load() % Int(11) == Int(0)).Then(Break()),
        For(
            i.store(n.load()),
            i.load() < Int(25),
            i.store(i.load() + Int(3)),
        ).Do(If(n.load() % Int(4) == Int(0)).Then(Continue()), Log(Itob(i.load()))),
        n.store(n.load() + i.load() - Int(1)),
    )


@Subroutine(TealType.uint64)  # type: ignore[misc]
def factorial(n: Expr) -> Expr:
    # Note: Recursion is not supported if the subroutine takes ScratchVar type argument
    return If(
        n == Int(0),  # condition
        Return(Int(1)),  # then expression
        Return(n * factorial(n - Int(1))),  # else expression
    )


def call_all_subroutines() -> Expr:
    n = ScratchVar()
    return Pop(
        Seq(
            [
                Pop(
                    factorial(Int(10))
                ),  # All Expressions in the Seq must return TealType.none. Last expression is a exception.
                foo_bar(Int(1)),
                n.store(Int(0)),
                break_continue(n),  # ScratchVar types are taken as references.
                factorial(Int(5)),  # Last expression can return a value
            ]
        )  # Seq returns value of the last expression. Pop it to return none.
    )


def some_subroutines(c: Expr) -> Expr:
    n = ScratchVar()
    return Cond(
        [c == Int(0), Pop(factorial(Int(7)))],  # All branches should return same type
        [
            c == Int(1),
            Seq(
                n.store(Int(1)),
                foo_bar(n.load()),
            ),
        ],
        [c == Int(2), Seq([n.store(Int(0)), break_continue(n)])],
    )


def approval_program() -> Expr:
    return Seq(
        [
            call_all_subroutines(),
            some_subroutines(Int(1)),
            Approve(),
        ]
    )


control_flow_ap = compileTeal(
    approval_program(),
    mode=Mode.Application,
    version=7,
    optimize=OptimizeOptions(scratch_slots=True),
)

if __name__ == "__main__":
    print(control_flow_ap)



from typing import Tuple, Dict
import pytest

from tealer.teal.subroutine import Subroutine
from tealer.teal.instructions import instructions
from tealer.teal.parse_teal import parse_teal

from tests.utils import cmp_cfg, construct_cfg


MULTIPLE_RETSUB = """
#pragma version 5
b main
push_zero:
    int 0
    retsub
is_even:
    int 2
    %
    bz return_1
    callsub push_zero
    retsub
return_1:
    int 1
    retsub
main:
    int 4
    callsub is_even
    return
"""

ins_list = [
    instructions.Pragma(5),
    instructions.B("main"),
    instructions.Label("push_zero"),
    instructions.Int(0),
    instructions.Retsub(),
    instructions.Label("is_even"),
    instructions.Int(2),
    instructions.Modulo(),
    instructions.BZ("return_1"),
    instructions.Callsub("push_zero"),
    instructions.Retsub(),
    instructions.Label("return_1"),
    instructions.Int(1),
    instructions.Retsub(),
    instructions.Label("main"),
    instructions.Int(4),
    instructions.Callsub("is_even"),
    instructions.Return(),
]

ins_partitions = [(0, 2), (2, 5), (5, 9), (9, 10), (10, 11), (11, 14), (14, 17), (17, 18)]
bbs_links = [(0, 6), (2, 3), (2, 5), (3, 4), (6, 7)]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)
MULTIPLE_RETSUB_MAIN = Subroutine("", bbs[0], [bbs[0], bbs[6], bbs[7]])
MULTIPLE_RETSUB_SUBROUTINES = {
    "push_zero": Subroutine("push_zero", bbs[1], [bbs[1]]),
    "is_even": Subroutine("is_even", bbs[2], [bbs[2], bbs[3], bbs[4], bbs[5]]),
}

SUBROUTINE_BACK_JUMP = """
#pragma version 5
b main
getmod:
    %
    retsub
is_odd:
    int 2
    b getmod
main:
    int 5
    callsub is_odd
    return
"""

ins_list = [
    instructions.Pragma(5),
    instructions.B("main"),
    instructions.Label("getmod"),
    instructions.Modulo(),
    instructions.Retsub(),
    instructions.Label("is_odd"),
    instructions.Int(2),
    instructions.B("getmod"),
    instructions.Label("main"),
    instructions.Int(5),
    instructions.Callsub("is_odd"),
    instructions.Return(),
]

ins_partitions = [(0, 2), (2, 5), (5, 8), (8, 11), (11, 12)]
bbs_links = [(0, 3), (3, 4), (2, 1)]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)
SUBROUTINE_BACK_JUMP_MAIN = Subroutine("", bbs[0], [bbs[0], bbs[3], bbs[4]])
SUBROUTINE_BACK_JUMP_SUBROUTINES = {"is_odd": Subroutine("is_odd", bbs[2], [bbs[1], bbs[2]])}

ALL_TESTS = [
    (MULTIPLE_RETSUB, MULTIPLE_RETSUB_MAIN, MULTIPLE_RETSUB_SUBROUTINES),
    (SUBROUTINE_BACK_JUMP, SUBROUTINE_BACK_JUMP_MAIN, SUBROUTINE_BACK_JUMP_SUBROUTINES),
]


@pytest.mark.parametrize("test", ALL_TESTS)  # type: ignore
def test_subroutine_identification(test: Tuple[str, Subroutine, Dict[str, Subroutine]]) -> None:
    code, expected_main, expected_subroutines = test
    teal = parse_teal(code.strip())
    test_main = teal.main

    assert cmp_cfg([test_main.entry], [expected_main.entry])
    assert cmp_cfg(test_main.blocks, expected_main.blocks)

    subroutines = teal.subroutines
    assert len(subroutines.keys()) == len(expected_subroutines.keys())
    assert sorted(subroutines.keys()) == sorted(expected_subroutines.keys())
    for ex_name in expected_subroutines:
        subroutine = subroutines[ex_name]
        expected_subroutine = expected_subroutines[ex_name]
        assert subroutine.name == expected_subroutine.name
        assert cmp_cfg([subroutine.entry], [expected_subroutine.entry])
        assert cmp_cfg(subroutine.blocks, expected_subroutine.blocks)
    assert len(subroutines) == len(expected_subroutines)



"""Detector for finding execution paths missing GroupSize check."""

from typing import List, TYPE_CHECKING, Tuple

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.teal.basic_blocks import BasicBlock
from tealer.teal.instructions.instructions import (
    Gtxn,
    Gtxna,
    Gtxnas,
    Gtxns,
    Gtxnsa,
    Gtxnsas,
)

from tealer.utils.algorand_constants import MAX_GROUP_SIZE
from tealer.utils.analyses import is_int_push_ins
from tealer.analyses.utils.stack_ast_builder import construct_stack_ast, UnknownStackValue
from tealer.detectors.utils import detect_missing_tx_field_validations_group
from tealer.utils.output import ExecutionPaths


if TYPE_CHECKING:
    from tealer.teal.teal import Teal
    from tealer.utils.output import ListOutput
    from tealer.teal.instructions.instructions import Instruction
    from tealer.teal.context.block_transaction_context import BlockTransactionContext


class MissingGroupSize(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find execution paths using absoulte index without checking group size."""

    NAME = "group-size-check"
    DESCRIPTION = "Usage of absolute indexes without validating GroupSize"
    TYPE = DetectorType.STATELESS_AND_STATEFULL

    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = (
        "https://github.com/crytic/tealer/wiki/Detector-Documentation#missing-groupsize-validation"
    )
    WIKI_TITLE = "Missing GroupSize Validation"
    WIKI_DESCRIPTION = (
        "Contract's execution depends on multiple transactions in the group"
        " and it uses absolute index to access information of other transactions in the group."
        " Attacker can exploit the contract by abusing the lack of validations on `GroupSize`."
        " More at [building-secure-contracts/not-so-smart-contracts/algorand/group_size_check]"
        "(https://github.com/crytic/building-secure-contracts/tree/master/not-so-smart-contracts/algorand/group_size_check)"
    )
    WIKI_EXPLOIT_SCENARIO = """
```py
def mint_wrapped_algo() -> Expr:
    validations = Assert(
        And(
            Gtxn[0].receiver() == Global.current_application_address(),
            Gtxn[0].type_enum() == TxnType.Payment,
        )
    )
    transfer_op = transfer_wrapped_algo(Txn.sender(), Gtxn[0].amount())
    return Seq([validations, transfer_op, Approve()])
```

Eve sends the following group transaction:
```
0. Payment of 1 million ALGOs to application address
1. Call mint_wrapped_algo
2. Call mint_wrapped_algo 
...
15. Call mint_wrapped_algo
```

Eve receives 15 million wrapped-algos instead of 1 million wrapped-algos.\
 Eve exchanges the Wrapped-algo to ALGO and steals 14 million ALGOs.
"""
    WIKI_RECOMMENDATION = """
- Avoid using absolute indexes. Validate GroupSize if used.
- Favor using ARC-4 ABI and relative indexes for group transactions.
"""

    @staticmethod
    def _accessed_using_absolute_index(bb: BasicBlock) -> bool:
        """Return True if a instruction in bb access a field using absolute index

        a. gtxn t f, gtxna t f i, gtxnas t f,
        b. gtxns f, gtxnsa f i, gtxnsas f

        Instructions in (a) take transaction index as a immediate argument.
        Return True if bb contains any one of those instructions.

        Instructions in (b) take transaction index from the stack.
        `gtxns f` and `gtxnsa f i` take only one argument and it is the transaction index.
        `gtxnsas f` takes two arguments and transaction index is the first argument.
        Return True if the transaction index is pushed by an int instruction.

        Args:
            bb: A basic block of the CFG

        Returns:
            Returns True if a instruction in bb access a field of group transaction using
            an absolute index. returns False otherwise.
        """
        stack_gtxns_ins: List["Instruction"] = []
        for ins in bb.instructions:
            if isinstance(ins, (Gtxn, Gtxna, Gtxnas)):
                return True
            if isinstance(ins, (Gtxns, Gtxnsa, Gtxnsas)):
                stack_gtxns_ins.append(ins)
        if not stack_gtxns_ins:
            return False
        ast_values = construct_stack_ast(bb)
        for ins in stack_gtxns_ins:
            index_value = ast_values[ins].args[0]
            if isinstance(index_value, UnknownStackValue):
                continue
            is_int, _ = is_int_push_ins(index_value.instruction)
            if is_int:
                return True
        return False

    def detect(self) -> "ListOutput":
        """Detect execution paths with missing GroupSize check.

        Returns:
            ExecutionPaths instance containing the list of vulnerable execution
            paths along with name, check, impact, confidence and other detector
            information.
        """

        def checks_group_size(block_ctx: "BlockTransactionContext") -> bool:
            # return True if group-size is checked in the path
            # otherwise return False
            if block_ctx.is_gtxn_context:
                # gtxn_contexts have group-sizes, group-indices set to 0.
                return False
            return MAX_GROUP_SIZE not in block_ctx.group_sizes

        def satisfies_report_condition(path: List["BasicBlock"]) -> bool:

            for block in path:
                if self._accessed_using_absolute_index(block):
                    return True
            return False

        output: List[
            Tuple["Teal", List[List["BasicBlock"]]]
        ] = detect_missing_tx_field_validations_group(
            self.tealer, checks_group_size, satisfies_report_condition
        )
        construct_stack_ast.cache_clear()

        detector_output: "ListOutput" = []
        for contract, vulnerable_paths in output:
            detector_output.append(ExecutionPaths(contract, self, vulnerable_paths))

        return detector_output



MAX_GROUP_SIZE = 16
MIN_ALGORAND_FEE = 1000  # in micro algos
ZERO_ADDRESS = "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEVAL4QAJS7JHB4"
MAX_UINT64 = (1 << 64) - 1  # 2**64 - 1
# Maximum number of inner transactions a group transaction can have.
MAX_NUM_INNER_TXN = 256

# Over estimation of maximum cost a transaction can consume.
# There can be a maximum of 256 inner transactions in a group.
# There can be maximum of 16 transactions in a group.
# Each inner txn and txn costs MIN_ALGORAND Fee
MAX_TRANSACTION_COST = (MAX_GROUP_SIZE + MAX_NUM_INNER_TXN) * MIN_ALGORAND_FEE



from typing import Type, Tuple
import base64

import pytest

from tealer.teal.instructions import instructions
from tealer.teal.instructions.instructions import (
    IntcInstruction,
    BytecInstruction,
)
from tealer.teal.instructions import transaction_field
from tealer.teal.global_field import GlobalField
from tealer.teal.instructions.transaction_field import TransactionField
from tealer.teal.instructions.acct_params_field import AcctParamsField
from tealer.teal.instructions.app_params_field import AppParamsField
from tealer.teal.instructions.asset_holding_field import AssetHoldingField
from tealer.teal.instructions.asset_params_field import AssetParamsField
from tealer.teal.instructions.parse_instruction import parse_line, ParseError
from tealer.teal.parse_teal import parse_teal
from tealer.teal.parse_functions import copy_main_cfg
from tealer.utils.analyses import is_int_push_ins, is_byte_push_ins
from tealer.exceptions import TealerException

from tests.utils import cmp_cfg

TARGETS = [
    "tests/parsing/teal1-instructions.teal",
    "tests/parsing/transaction_fields.teal",
    "tests/parsing/global_fields.teal",
    "tests/parsing/instructions.teal",
    "tests/parsing/instructions.teal",
    "tests/parsing/instructions.teal",
    "tests/parsing/teal3-gload.teal",
    "tests/parsing/teal4-test0.teal",
    "tests/parsing/teal4-test1.teal",
    "tests/parsing/teal4-test2.teal",
    "tests/parsing/teal4-test3.teal",
    "tests/parsing/teal4-test4.teal",
    "tests/parsing/teal4-test5.teal",
    "tests/parsing/teal4-random-opcodes.teal",
    "tests/parsing/comments.teal",
    "tests/parsing/asset_holding_get.teal",
    "tests/parsing/asset_params_get.teal",
    "tests/parsing/teal5-app_params_get.teal",
    "tests/parsing/teal5-ecdsa.teal",
    "tests/parsing/teal5-extract.teal",
    "tests/parsing/teal5-itxn.teal",
    "tests/parsing/teal5-itxna.teal",
    "tests/parsing/teal5-mixed.teal",
    "tests/parsing/teal-instructions-with-versions.teal",
    "tests/parsing/teal-fields-with-versions.teal",
    "tests/parsing/multiple_retsub.teal",
    "tests/parsing/subroutine_jump_back.teal",
    "tests/parsing/teal6-acct_params_get.teal",
    "tests/parsing/teal6-instructions.teal",
    "tests/parsing/teal7-instructions.teal",
    "tests/parsing/teal8-instructions.teal",
]

TEST_CODE = """
intcblock 0xf 017 15
intcblock
int pay
pushint 1
byte base64 AA
byte b64 AA
byte base64(AA)
byte b64(AA)
byte base32 AA
byte b32 AA
byte base32(AA)
byte b32(AA)
byte 0x0123456789abcdef
byte "\x01\x02"
byte "string literal"
bytecblock b32(AA) base64 AA 0x00 "00"
bytecblock
byte "not label: // not comment either"
labelwithqoute": // valid
gtxn 1 Sender
gtxna 1 Applications 0
extract 0 1
gtxnas 1 Applications
gitxn 1 Sender
replace 1
replace2 1
match label1 label2
switch label1 label2
label1:
int 1
label2:
int 2
method "add(uint64)uint64"
"""

invalid_instructions = """
byte "sdf       // missing closing qoute
byte base64     // missing encoded string
byte b64        // missing encoded string
byte base32     // missing encoded string
byte b32        // missing encoded string
byte base64(AA  // missing closing paranthesis
byte b64(AA     // missing closing paranthesis
byte base32(AA  // missing closing paranthesis
byte b32(AA     // missing closing paranthesis
label: add      // contains additional token add
byte "1" "2"    // byte expects only one immediate arg.
byte sdf        // incorrect byte format
"""

unsupported_instructions = """
not an instruction
"""


@pytest.mark.parametrize("target", TARGETS)  # type: ignore
def test_parsing(target: str) -> None:
    with open(target, encoding="utf-8") as f:
        teal = parse_teal(f.read())
    # print instruction to trigger __str__ on each ins
    # print stack pop/push to trigger the function on each ins
    for i in teal.instructions:
        assert not isinstance(i, instructions.UnsupportedInstruction), f'ins "{i}" is not supported'
        print(i, i.cost)
        print(i, i.stack_pop_size)
        print(i, i.stack_push_size)


# pylint: disable=too-many-locals
@pytest.mark.parametrize("target", TARGETS)  # type: ignore
def test_copy_main_cfg(target: str) -> None:
    with open(target, encoding="utf-8") as f:
        teal = parse_teal(f.read())
    copied_main = copy_main_cfg(teal)

    assert cmp_cfg(copied_main, teal.main.blocks)
    copied_blocks = sorted(copied_main, key=lambda bi: bi.idx)
    original_blocks = sorted(teal.main.blocks, key=lambda bi: bi.idx)
    copied_instructions = [ins for bi in copied_blocks for ins in bi.instructions]
    # original_instructions = [ins for bi in original_blocks for ins in bi.instructions]

    print(vars(copied_blocks[0]).keys())
    print(vars(copied_instructions[0]).keys())

    ignore_block_fields = ["_instructions", "_prev", "_next", "_transaction_context", "_subroutine"]
    ignore_instruction_fields = ["_prev", "_next", "_bb"]
    for bb_copy, bb_orig in zip(copied_blocks, original_blocks):
        for key in vars(bb_copy):
            if key in ignore_block_fields:
                continue
            print(key)
            assert vars(bb_copy)[key] == vars(bb_orig)[key]
        for ins_copy, ins_orig in zip(bb_copy.instructions, bb_orig.instructions):
            ins_copy_vars = vars(ins_copy)
            ins_orig_vars = vars(ins_orig)
            for key in ins_copy_vars:
                if key in ignore_instruction_fields:
                    continue
                if isinstance(
                    ins_copy_vars[key],
                    (
                        TransactionField,
                        GlobalField,
                        AppParamsField,
                        AcctParamsField,
                        AssetParamsField,
                        AssetHoldingField,
                    ),
                ):
                    continue
                print(key)
                assert ins_copy_vars[key] == ins_orig_vars[key]


def _cmp_instructions(
    b1: instructions.Instruction,
    b2: instructions.Instruction,
    target: Type[instructions.Instruction],
    attributes: Tuple[str],
) -> bool:
    check = isinstance(b1, target) and isinstance(b2, target)
    if not check:
        return check

    for attr in attributes:
        v1 = getattr(b1, attr, None)
        v2 = getattr(b2, attr, None)
        if v1 != v2:
            return False
    return True


def test_parsing_2() -> None:
    teal = parse_teal(TEST_CODE)
    ins1 = teal.instructions
    ins2 = [
        instructions.Intcblock([15, 15, 15]),
        instructions.Intcblock([]),
        instructions.Int("pay"),
        instructions.PushInt(1),
        instructions.Byte("0x00"),
        instructions.Byte("0x00"),
        instructions.Byte("0x00"),
        instructions.Byte("0x00"),
        instructions.Byte("0x00"),
        instructions.Byte("0x00"),
        instructions.Byte("0x00"),
        instructions.Byte("0x00"),
        instructions.Byte("0x0123456789abcdef"),
        instructions.Byte('"\x01\x02"'),
        instructions.Byte('"string literal"'),
        instructions.Bytecblock(["0x00", "0x00", "0x00", '"00"']),
        instructions.Bytecblock([]),
        instructions.Byte('"not label: // not comment either"'),
        instructions.Label('labelwithqoute"'),
        instructions.Gtxn(1, transaction_field.Sender()),
        instructions.Gtxna(1, transaction_field.Applications(0)),
        instructions.Extract(0, 1),
        instructions.Gtxnas(1, transaction_field.Applications(-1)),
        instructions.Gitxn(1, transaction_field.Sender()),
        instructions.Replace(1),
        instructions.Replace2(1),
        instructions.Match(["label1", "label2"]),
        instructions.Switch(["label1", "label2"]),
        instructions.Label("label1"),
        instructions.Int(1),
        instructions.Label("label2"),
        instructions.Int(2),
        instructions.Method('"add(uint64)uint64"'),
    ]
    t = [
        (instructions.Intcblock, ("_constants",)),
        (instructions.Intcblock, ("_constants",)),
        (instructions.Int, ("value",)),
        (instructions.PushInt, ("value",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Bytecblock, ("_constants",)),
        (instructions.Bytecblock, ("_constants",)),
        (instructions.Byte, ("_bytes",)),
        (instructions.Label, ("label",)),
        (instructions.Gtxn, ("idx",)),
        (instructions.Gtxna, ("idx",)),
        (instructions.Extract, ("start_position", "length")),
        (instructions.Gtxnas, ("idx",)),
        (instructions.Gitxn, ("idx",)),
        (instructions.Replace, ("start_position", "is_replace2", "is_replace3")),
        (instructions.Replace2, ("start_position",)),
        (instructions.Match, ("labels",)),
        (instructions.Switch, ("labels",)),
        (instructions.Label, ("label",)),
        (instructions.Int, ("value",)),
        (instructions.Label, ("label",)),
        (instructions.Int, ("value",)),
        (instructions.Method, ("method_signature",)),
    ]

    attributes: Tuple[str]
    for (b1, b2, (target, attributes)) in zip(ins1, ins2, t):  # type: ignore
        assert _cmp_instructions(b1, b2, target, attributes)


def test_invalid_instructions() -> None:
    for line in invalid_instructions.strip().splitlines():
        with pytest.raises(ParseError):
            parse_line(line)


def test_unsupported_instructions() -> None:
    for line in unsupported_instructions.strip().splitlines():
        ins = parse_line(line)
        assert isinstance(ins, instructions.UnsupportedInstruction)
        assert str(ins) == f"UNSUPPORTED {line.strip()}"
        assert ins.verbatim_line == line.strip()


def test_field_properties() -> None:
    ins = parse_line("gitxnas 1 ApplicationArgs")
    assert (
        isinstance(ins, instructions.Gitxnas)
        and ins.idx == 1
        and isinstance(ins.field, transaction_field.TransactionArrayField)
        and ins.field.idx == -1
    )

    ins = parse_line("gitxna 1 ApplicationArgs 0")
    assert (
        isinstance(ins, instructions.Gitxna)
        and ins.idx == 1
        and isinstance(ins.field, transaction_field.TransactionArrayField)
        and ins.field.idx == 0
    )


def test_instruction_properties() -> None:
    CURRENT_TEST_CODE = """
    int 1 // comment
    int 2
    """
    teal = parse_teal(CURRENT_TEST_CODE)
    ins1 = teal.instructions[0]
    ins2 = teal.instructions[1]
    assert ins1.prev == []
    assert ins1.next == [ins2]
    assert ins2.prev == [ins1]
    assert ins2.next == []
    assert ins1.comment == "// comment"

    # accessing replace instruction start_position fails if it is None i.e if there's no immediate argument.
    # it should be checked that whether given replace instruction is semantically equivalent to replace2 or replace3.
    ins = parse_line("replace 1")
    assert isinstance(ins, instructions.Replace) and ins.is_replace2 and ins.start_position == 1

    ins = parse_line("replace")
    assert isinstance(ins, instructions.Replace) and ins.is_replace3
    with pytest.raises(Exception):
        print(ins.start_position)


def test_cost_values() -> None:
    CURRENT_TEST_CODE = """
    sha256
    sha512_256
    keccak256
    ecdsa_verify Secp256k1
    ecdsa_pk_decompress Secp256k1
    ecdsa_pk_recover Secp256k1
    b%
    b&
    b|
    b+
    b-
    b/
    b*
    b^
    b~
    divmodw
    expw
    sqrt
    bsqrt
    base64_decode URLEncoding
    json_ref JSONUint64
    ed25519verify_bare
    vrf_verify VrfAlgorand
    """
    for line in CURRENT_TEST_CODE.strip().splitlines():
        # when cost parameter accessed, it checks that instruction object's BasicBlock is not none. if it is none, then
        # cost property raises exception. These tests are included to cover those brances.
        with pytest.raises(TealerException):
            print(line)
            # pylint: disable=expression-not-assigned
            parse_line(line).cost  # type: ignore

    # cost should return 0 if the contract version is less than instruction supported version
    CURRENT_TEST_CODE = """
    ecdsa_verify Secp256k1
    ecdsa_pk_decompress Secp256k1
    ecdsa_pk_recover Secp256k1
    bsqrt
    base64_decode URLEncoding
    json_ref JSONUint64
    ed25519verify_bare
    """
    teal = parse_teal(CURRENT_TEST_CODE)
    for ins in teal.instructions:
        print("DSDF", ins, ins.cost)
        assert ins.cost == 0


INTCBLOCK_TESTS = [
    "tests/parsing/intcblock_1.teal",
    "tests/parsing/intcblock_2.teal",
    "tests/parsing/intcblock_3.teal",
]


@pytest.mark.parametrize("test", INTCBLOCK_TESTS)  # type: ignore
def test_intc_bytec(test: str) -> None:
    with open(test, encoding="utf-8") as f:
        teal = parse_teal(f.read())

    for ins in teal.instructions:
        if isinstance(ins, IntcInstruction):
            is_known, value = is_int_push_ins(ins)
            assert is_known
            assert value == int(ins.comment[2:].strip())

        if isinstance(ins, BytecInstruction):
            is_known, value = is_byte_push_ins(ins)
            assert is_known
            expected_value = ins.comment[2:].strip()
            if expected_value.startswith('"'):
                expected_value = "0x" + expected_value[1:-1].encode().hex()
            elif expected_value.startswith("addr"):
                s = expected_value.strip("addr ").strip()
                expected_value = (
                    "0x" + base64.b32decode(s + "=" * (-len(s) % 8))[:-4].hex()
                )  # 4-byte checksum
            assert value == expected_value


INTCBLOCK_FALSE_TEST_1 = """
#pragma version 7
intcblock 0x00 0x01 0x02 0x03 0x04 0x05 0x06
intcblock 0x07 0x08 0x09 0x10 0x11 0x12 0x13
intc0
intc1
intc2
intc3
intc 4
return
"""

INTCBLOCK_FALSE_TEST_2 = """
#pragma version 7
int 0
int 1
+
b next
next:
intcblock 0x00 0x01 0x02 0x03 0x04 0x05 0x06
intc0
intc1
intc2
intc3
intc 4
return
"""

INTCBLOCK_FALSE_TEST_3 = """
#pragma version 7
intc0
intc1
return
"""

# Tealer can determine that this is invalid code in this case. But currently does
# not do that to be uniform for all cases.
INTCBLOCK_FALSE_TEST_4 = """
#pragma version 7
intcblock 0x00 0x01 0x02 0x03
intc 7  // Invalid index
"""

# Tealer cannot determing in this case.
INTCBLOCK_FALSE_TEST_5 = """
#pragma version 7
intcblock 0x00 0x01 0x02 0x03
txn RekeyTo
global ZeroAddress
==
bnz branch
b sub
branch:
    intcblock 0x04 0x05 0x06 0x07 0x08 0x09 0xa
sub:
    intc 9 // Invalid index
"""

INTCBLOCK_FALSE_TESTS = [
    INTCBLOCK_FALSE_TEST_1,
    INTCBLOCK_FALSE_TEST_2,
    INTCBLOCK_FALSE_TEST_3,
    INTCBLOCK_FALSE_TEST_4,
    INTCBLOCK_FALSE_TEST_5,
]


@pytest.mark.parametrize("test", INTCBLOCK_FALSE_TESTS)  # type: ignore
def test_intc_bytec_false(test: str) -> None:
    teal = parse_teal(test)
    for ins in teal.instructions:
        if isinstance(ins, IntcInstruction):
            is_int, value = is_int_push_ins(ins)
            assert is_int and value is None

    teal = parse_teal(test.replace("intc", "bytec"))
    for ins in teal.instructions:
        if isinstance(ins, BytecInstruction):
            is_bytes, value = is_byte_push_ins(ins)
            assert is_bytes and value is None



from Cryptodome.Hash import SHA512


def get_method_selector(method_signature: str) -> str:
    hash_obj = SHA512.new(truncate="256")
    hash_obj.update(method_signature.encode("utf-8"))
    return f"0x{hash_obj.hexdigest()[:8]}"



"""Contains functions, classes for parsing and representing teal instructions.

For representing teal code, this module contains classes for teal
instructions, transaction fields, App parameter fields, Asset parameter
fields and Asset holding fields.

This module also contains implementations of parsers for instructions,
global fields, transaction fields, App parameter fields,
Asset parameter fields and Asset holding fields.

Modules:
    app_params_field: Defines classes to represent app parameter fields.

    asset_holding_field: Defines classes to represent asset holding fields.

    asset_params_field: Defines classes to represent asset parameter fields.

    instructions: Defines classes to represent teal instructions.

    parse_app_params_field: Parser for app parameter fields.

    parse_asset_holding_field: Parser for asset holding fields.

    parse_asset_params_field: Parser for asset parameter fields.

    parse_global_field: Parser for global fields.

    parse_instruction: Parser for teal instructions.

    parse_transaction_field: Parser for transaction fields.

    transaction_field: Defines classes to represent transaction fields.
"""



from typing import List

from tealer.detectors.all_detectors import IsUpdatable


MULTIPLE_RETURN_POINTS = """
#pragma version 7       // B0
callsub subroutine_0
callsub subroutine_0    // B1
int 1                   // B2
return
subroutine_0:           // B3
    retsub
"""

# There is only one possible execution path B0 -> B3 -> B1 -> B3 -> B2
# 1.) In the CFG B3(subroutine_0) is connected to B1 and B2. And when traversing the CFG, detector
# considers B0 -> B3 -> B2 as a valid path and reports it.
#
# 2.) Path B0 -> B3 -> B1 -> B3 -> B2 is not reported because `subroutine_0` is called twice and
# on the second call, detectors think that they are in a loop because subroutine's blocks are
# are already in the path(after the first call).

# Before fixing bug (1) and (2)
MULTIPLE_RETURN_POINTS_VULNERABLE_PATHS_1: List[List[int]] = [
    [0, 3, 2],
]

# After fixing bug (1) and not (2)
MULTIPLE_RETURN_POINTS_VULNERABLE_PATHS_2: List[List[int]] = []

# After fixing bug (1) and (2)
MULTIPLE_RETURN_POINTS_VULNERABLE_PATHS_3: List[List[int]] = [
    [0, 3, 1, 3, 2],
]

# Tealer does not identify a subroutine when `callsub` instruction is the last instruction.
# Bug is fixed now
CALLSUB_AT_END = """
#pragma version 7
b main
sub:
    int 1
    return
main:
    callsub sub
"""

CALLSUB_AT_END_VULNERABLE_PATHS: List[List[int]] = [
    [0, 2, 1],
]

subroutine_patterns_tests = [
    # (MULTIPLE_RETURN_POINTS, IsUpdatable, MULTIPLE_RETURN_POINTS_VULNERABLE_PATHS_1),
    # (MULTIPLE_RETURN_POINTS, IsUpdatable, MULTIPLE_RETURN_POINTS_VULNERABLE_PATHS_2),
    (MULTIPLE_RETURN_POINTS, IsUpdatable, MULTIPLE_RETURN_POINTS_VULNERABLE_PATHS_3),
    (CALLSUB_AT_END, IsUpdatable, CALLSUB_AT_END_VULNERABLE_PATHS),
]



from typing import TYPE_CHECKING, List, Set, Tuple, Dict

from tealer.analyses.dataflow.transaction_context.generic import DataflowTransactionContext
from tealer.teal.instructions.instructions import (
    Global,
    Eq,
    Neq,
    Greater,
    GreaterE,
    Less,
    LessE,
    Txn,
)
from tealer.teal.global_field import GroupSize
from tealer.teal.instructions.transaction_field import (
    GroupIndex,
)
from tealer.utils.analyses import is_int_push_ins
from tealer.utils.algorand_constants import MAX_GROUP_SIZE
from tealer.analyses.utils.stack_ast_builder import KnownStackValue, UnknownStackValue

if TYPE_CHECKING:
    from tealer.teal.instructions.instructions import Instruction
    from tealer.teal.basic_blocks import BasicBlock

# TODO: Change GroupSize, GroupIndex values representation to something similar of FeeValue.
group_size_key = "GroupSize"
group_index_key = "GroupIndex"
analysis_keys = [group_size_key, group_index_key]
universal_sets = {}
universal_sets[group_size_key] = list(range(1, MAX_GROUP_SIZE + 1))
universal_sets[group_index_key] = list(range(0, MAX_GROUP_SIZE))


class GroupIndices(DataflowTransactionContext):  # pylint: disable=too-few-public-methods

    GROUP_SIZE_KEY = group_size_key
    GROUP_INDEX_KEY = group_index_key
    BASE_KEYS: List[str] = analysis_keys
    KEYS_WITH_GTXN: List[str] = []  # gtxn information is not collected for any of the keys
    UNIVERSAL_SETS: Dict[str, List] = universal_sets

    def _universal_set(self, key: str) -> Set:
        return set(self.UNIVERSAL_SETS[key])

    def _null_set(self, key: str) -> Set:
        return set()

    def _union(self, key: str, a: Set, b: Set) -> Set:
        return a | b

    def _intersection(self, key: str, a: Set, b: Set) -> Set:
        return a & b

    @staticmethod
    def _get_asserted_int_values(
        comparison_ins: "Instruction", compared_int: int, universal_set: List[int]
    ) -> List[int]:
        """return list of ints from universal set(U) that will satisfy the comparison.

        if the given condition uses ==, return compared int list.
        if the condition uses != then return {U - compared_int}
        if the given condition uses <, return U[ : U.index(compared_int)]
        if the given condition uses <=, return U[ : U.index(compared_int) + 1]
        if the given condition uses >, return U[U.index(compared_int) + 1:]
        if the given condition uses >=, return U[U.index(compared_int): ]

        Args:
            comparison_ins: comparison instruction used. can be [==, !=, <, <=, >, >=]
            compared_int: integer value compared.
            universal_set: list of all possible integer values for the field.

        Returns:
            list of ints that will satisfy the comparison
        """
        U = list(universal_set)

        if isinstance(comparison_ins, Eq):  # pylint: disable=no-else-return
            return [compared_int]
        elif isinstance(comparison_ins, Neq):
            if compared_int in U:
                U.remove(compared_int)
            return U
        elif isinstance(comparison_ins, Less):
            return [i for i in U if i < compared_int]
        elif isinstance(comparison_ins, LessE):
            return [i for i in U if i <= compared_int]
        elif isinstance(comparison_ins, Greater):
            return [i for i in U if i > compared_int]
        elif isinstance(comparison_ins, GreaterE):
            return [i for i in U if i >= compared_int]
        else:
            return U

    def _get_asserted_groupsizes(
        self, ins_stack_value: KnownStackValue
    ) -> Tuple[Set[int], Set[int]]:
        """return set of values for groupsize that will make the comparison true and false

        checks for instruction sequence and returns group size values that will make the comparison true.

        [ Global GroupSize | (int | pushint)]
        [ (int | pushint) | Global GroupSize]
        [ == | != | < | <= | > | >=]

        Args:
            ins_stack_value: The stack value. This value represents the result of the group-size check.

        Returns:
            set of groupsize values that will make the comparison true, set of groupsize values that will make the comparison false.
        """
        U = list(self.UNIVERSAL_SETS[self.GROUP_SIZE_KEY])

        if isinstance(ins_stack_value.instruction, (Eq, Neq, Less, LessE, Greater, GreaterE)):
            arg1 = ins_stack_value.args[0]
            arg2 = ins_stack_value.args[1]
            if isinstance(arg1, UnknownStackValue) or isinstance(arg2, UnknownStackValue):
                # if one of the args is unknown
                return set(U), set(U)
            ins1 = arg1.instruction
            ins2 = arg2.instruction
            compared_value = None

            if isinstance(ins1, Global) and isinstance(ins1.field, GroupSize):
                is_int, value = is_int_push_ins(ins2)
                if is_int:
                    compared_value = value
            elif isinstance(ins2, Global) and isinstance(ins2.field, GroupSize):
                is_int, value = is_int_push_ins(ins1)
                if is_int:
                    compared_value = value

            if compared_value is None or not isinstance(compared_value, int):
                # if the comparison does not check groupsize, return U as values that make the comparison false
                return set(U), set(U)

            ins = ins_stack_value.instruction
            asserted_values = self._get_asserted_int_values(ins, compared_value, U)
            return set(asserted_values), set(U) - set(asserted_values)
        return set(U), set(U)

    def _get_asserted_groupindices(
        self, ins_stack_value: KnownStackValue
    ) -> Tuple[Set[int], Set[int]]:
        """return list of values for group index that will make the comparison true and false

        checks for instruction sequence and returns group index values that will make the comparison true.

        [ txn GroupIndex | (int | pushint)]
        [ (int | pushint) | txn GroupIndex]
        [ == | != | < | <= | > | >=]

        Args:
            ins_stack_value: The stack value. This value represents the result of the group index check.

        Returns:
            List of groupindex values that will make the comparison true.
        """
        U = list(self.UNIVERSAL_SETS[self.GROUP_INDEX_KEY])

        if isinstance(ins_stack_value.instruction, (Eq, Neq, Less, LessE, Greater, GreaterE)):
            arg1 = ins_stack_value.args[0]
            arg2 = ins_stack_value.args[1]
            if isinstance(arg1, UnknownStackValue) or isinstance(arg2, UnknownStackValue):
                # if one of the args is unknown
                return set(U), set(U)
            ins1 = arg1.instruction
            ins2 = arg2.instruction
            compared_value = None

            if isinstance(ins1, Txn) and isinstance(ins1.field, GroupIndex):
                is_int, value = is_int_push_ins(ins2)
                if is_int:
                    compared_value = value
            elif isinstance(ins2, Txn) and isinstance(ins2.field, GroupIndex):
                is_int, value = is_int_push_ins(ins1)
                if is_int:
                    compared_value = value

            if compared_value is None or not isinstance(compared_value, int):
                return set(U), set(U)

            ins = ins_stack_value.instruction
            asserted_values = self._get_asserted_int_values(ins, compared_value, U)
            return set(asserted_values), set(U) - set(asserted_values)
        return set(U), set(U)

    def _get_asserted_single(self, key: str, ins_stack_value: KnownStackValue) -> Tuple[Set, Set]:
        if key == self.GROUP_SIZE_KEY:
            return self._get_asserted_groupsizes(ins_stack_value)
        return self._get_asserted_groupindices(ins_stack_value)

    def _store_results(self) -> None:
        # use group_sizes to update group_indices
        group_sizes_context = self._block_contexts[self.GROUP_SIZE_KEY]
        group_indices_context = self._block_contexts[self.GROUP_INDEX_KEY]
        for bi in self._function.blocks:
            group_indices_context[bi] = group_indices_context[bi] & set(
                range(0, max(group_sizes_context[bi], default=0))
            )

        group_size_block_context = self._block_contexts[self.GROUP_SIZE_KEY]
        for block in self._function.blocks:
            self._function.transaction_context(block).group_sizes = list(
                group_size_block_context[block]
            )

        group_index_block_context = self._block_contexts[self.GROUP_INDEX_KEY]
        for block in self._function.blocks:
            self._function.transaction_context(block).group_indices = list(
                group_index_block_context[block]
            )
